from flask import current_app

class SummaryCharts:
	#
	# prepares jsons for samary charts
	#

	# status strings from GlobalConstants
	TESTCASESTATUS_SUCCESS = "OK"
	TESTCASESTATUS_ERROR = "Failed"
	TESTCASESTATUS_WAITING = "Paused"


	def __init__(self, log, min_length=10):
		self.log = log
		self.shortest = min_length

	def get_charts(self):
		#
		# summary set of charts
		#
		return {
			'figures': self.figures(),
			'status': self.status(),
			'testcases': self.testcases(),
			'files': self.files(),
		}

	def files(self):
		return [
			{
				'title': 'Logfile',
				'url': '/'.join((
					'http:/',
					current_app.config.get('BAANGT_DATAFILE_HOST'),
					current_app.config.get('BAANGT_DATAFILE_GET'),
					self.log['id'],
					current_app.config.get('BAANGT_DATAFILE_LOGFILE'),
				)),
			},
			{
				'title': 'Results',
				'url': '/'.join((
					'http:/',
					current_app.config.get('BAANGT_DATAFILE_HOST'),
					current_app.config.get('BAANGT_DATAFILE_GET'),
					self.log['id'],
					current_app.config.get('BAANGT_DATAFILE_RESULTS'),
				)),
			},
		]

	def figures(self):
		#
		# builds json with log statistics
		#

		return {
			'records': self.log['Summary'].get('TestRecords'),
			'successful': self.log['Summary'].get('Successful'),
			'error': self.log['Summary'].get('Error'),
			'paused': self.log['Summary'].get('Paused'),
		}

	def status(self):
		#
		# status chart
		#

		return {
			'type': 'doughnut',
			'data': {
				'datasets': [{
					'data': [
						self.log['Summary'].get('Successful'),
						self.log['Summary'].get('Error'),
						self.log['Summary'].get('Paused'),
					],
					'backgroundColor': [
						'#52ff52',
						'#ff5252',
						'#bdbdbd',
					],
				}],
				'labels': [
					'Passed',
					'Failed',
					'Paused'
				],
			},
			'options': {
				'legend': {
					'display': False,
				},
			},
		}

	def testcases(self):
		#
		# builds json for Chart.js: duration of testcases
		#

		testcases = []

		# collect data for every TestCaseSequence
		for tcs in self.log['TestSequences']:
			null = max(0, self.shortest - len(tcs['TestCases']))
			testcases.append({
				'type': 'bar',
					'data': {
						'datasets': [
							{
								'data': [self.duration_to_seconds(t['Parameters'].get('Duration')) \
									if t['Parameters'].get('TestCaseStatus') == self.TESTCASESTATUS_SUCCESS else 0 \
									for t in tcs['TestCases']] + [None]*null,
								'backgroundColor': '#52ff52',
								'label': 'PASSED',
							},
							{
								'data': [self.duration_to_seconds(t['Parameters'].get('Duration')) \
									if t['Parameters'].get('TestCaseStatus') == self.TESTCASESTATUS_ERROR else 0 \
									for t in tcs['TestCases']] + [None]*null,
								'backgroundColor': '#ff5252',
								'label': 'FAILED',
							},
							{
								'data': [self.duration_to_seconds(t['Parameters'].get('Duration')) \
									if t['Parameters'].get('TestCaseStatus') == self.TESTCASESTATUS_WAITING else 0 \
									for t in tcs['TestCases']] + [None]*null,
								'backgroundColor': '#bdbdbd',
								'label': 'PAUSED',
							},
						],
						'labels': [t['Parameters'].get('Duration') for t in tcs['TestCases']] + [None]*null,
					},
					'options': {
						'legend': {
							'display': False,
						},
						'tooltips': {
							'mode': 'index',
							'intersect': False,
						},
						'scales': {
							'xAxes': [{
								'gridLines': {
									'display': False,
									'drawBorder': False,
								},
								'ticks': {
									'display': False,
								},
								'stacked': True,
							}],
							'yAxes': [{
								'gridLines': {
									'display': False,
									'drawBorder': False,
								},
								'ticks': {
									'display': False,
								},
								'stacked': True,
							}],
						},
					},
				})

		return testcases
		

	def duration_to_seconds(self, duration):
		if duration:
			without_decimals = duration.split('.')[0]
			factors = [3600, 60, 1]
			
			return sum(t*f for t,f in zip(map(int, without_decimals.split(':')), factors))

		return 0


class DashboardCharts:
	#
	# prepares jsons for dashboard Charts.js
	#

	def __init__(self, logs, length=10):
		self.logs = logs[-length:]
		self.null = length - len(self.logs)


	def get_charts(self):
		#
		# dashboard chart set
		#

		return {
			'figures': self.figures(),
			'status': self.status(),
			'results': self.results(),
			'duration': self.duration(),
		}

	def figures(self):
		#
		# builds json with log statistics
		#

		return {
			'records': self.logs[-1].testrecords,
			'successful': self.logs[-1].successful,
			'error': self.logs[-1].error,
			'paused': self.logs[-1].paused,
		}

	def status(self):
		#
		# status chart
		#

		return {
			'type': 'doughnut',
			'data': {
				'datasets': [{
					'data': [
						self.logs[-1].successful,
						self.logs[-1].error,
						self.logs[-1].paused,
					],
					'backgroundColor': [
						'#52ff52',
						'#ff5252',
						'#bdbdbd',
					],
				}],
				'labels': [
					'Passed',
					'Failed',
					'Paused'
				],
			},
			'options': {
				'legend': {
					'display': False,
				},
			},
		}


	def results(self):
		#
		# results chart
		#

		return {
			'type': 'bar',
			'data': {
				'datasets': [
					{
						'data': [x.successful for x in self.logs] + [None]*self.null,
						'backgroundColor': '#52ff52',
						'label': 'Passed',
					},
					{
						'data': [x.error for x in self.logs] + [None]*self.null,
						'backgroundColor': '#ff5252',
						'label': 'Failed',
					},
					{
						'data': [x.paused for x in self.logs] + [None]*self.null,
						'backgroundColor': '#bdbdbd',
						'label': 'Paused',
					},
				],
				'labels': [x.created.strftime('%Y-%m-%d %H:%M:%S') for x in self.logs] + [None]*self.null,
			},
			'options': {
				'legend': {
					'display': False,
				},
				'tooltips': {
					'mode': 'index',
					'intersect': False,
				},
				'scales': {
					'xAxes': [{
						'gridLines': {
							'display': False,
							'drawBorder': False,
						},
						'ticks': {
							'display': False,
						},
						'stacked': True,
					}],
					'yAxes': [{
						'gridLines': {
							'display': False,
							'drawBorder': False,
						},
						'ticks': {
							'display': False,
						},
						'stacked': True,
					}],
				},
			},
		}

	def duration(self):
		#
		# duration chart
		#

		return {
			'type': 'line',
			'data': {
				'datasets': [{
					'data': [x.duration for x in self.logs] + [None]*self.null,
					'fill': False,
					'borderColor': '#cfd8dc',
					'borderWidth': 1, 
					'pointBackgroundColor': 'transparent',
					'pointBorderColor': '#0277bd',
					'pointRadius': 5,
					'lineTension': 0,
					'label': 'Duration',
				}],
				'labels': [x.created.strftime('%Y-%m-%d %H:%M:%S') for x in self.logs] + [None]*self.null,
			},
			'options': {
				'legend': {
					'display': False,
				},
				'tooltips': {
					'mode': 'index',
					'intersect': False,
				},
				'layout': {
					'padding': {
						'top': 10,
						'right': 10,
					},
				},
				'scales': {
					'xAxes': [{
						'gridLines': {
							'display': False,
							'drawBorder': False,
						},
						'ticks': {
							'display': False,
						},
					}],
					'yAxes': [{
						'gridLines': {
							'display': False,
							'drawBorder': False,
						},
						'ticks': {
							'display': False,
						},
					}],
				},
			},
		}