from flask_wtf import FlaskForm
from wtforms.fields import StringField, PasswordField, SelectMultipleField, SelectField, FileField, HiddenField
from wtforms.widgets import TextArea, ListWidget, CheckboxInput
from wtforms.validators import ValidationError, DataRequired, EqualTo
from app.models import User
from app import utils
import json


#
# authantication forms
#

class LoginForm(FlaskForm):
	username = StringField('Username', validators=[DataRequired()])
	password = PasswordField('Password', validators=[DataRequired()])

	def validate_username(self, username):
		if User.query.filter_by(username=username.data).count() == 0:
			raise ValidationError('Wrong username')

	def validate_password(self, password):
		user = User.query.filter_by(username=self.username.data).first()
		if user and not user.verify_password(password.data):
			raise ValidationError('Wrong password')

class SingupForm(FlaskForm):
	username = StringField('Username', validators=[DataRequired()])
	password = PasswordField('Password', validators=[DataRequired()])
	password2 = PasswordField('Password again', validators=[DataRequired(), EqualTo('password')])

	def validate_username(self, username):
		if User.query.filter_by(username=username.data).count() > 0:
			raise ValidationError('This username is in use - please try another.')

#
# testrun import forms
#

class TestrunImportForm(FlaskForm):

	file = FileField('Testrun Definition', validators=[DataRequired()])
	#
	# uncomment to have more datafiles for import
	#
	dataFile1 = FileField(f'Data File 1')
	dataFile2 = FileField(f'Data File 2')
	dataFile3 = FileField(f'Data File 3')
	#dataFile4 = FileField(f'Data File 4')
	#dataFile5 = FileField(f'Data File 5')

	@property
	def dataFiles(cls):
		return [
			cls.dataFile1,
			cls.dataFile2,
			cls.dataFile3,
			#cls.dataFile4,
			#cls.dataFile5,
		]

class DataFileUpdateForm(FlaskForm):

	datafile = FileField('Testrun Definition', validators=[DataRequired()])
	datafile_id = HiddenField()
	testcaseseq_id = HiddenField()


#
# testrun item create forms
#

class TestrunCreateForm(FlaskForm):
	name = StringField('Name', validators=[DataRequired()])
	description = StringField('Description', validators=[DataRequired()], widget=TextArea())
	testcase_sequences = SelectMultipleField('Test Case Sequences')

	# clobals
	global_settings = StringField('Global Settings', render_kw={"placeholder": "Leave blank for defaults"}, widget=TextArea())

	def validate_global_settings(self, global_settings):
		if global_settings.data:
			try:
				json.loads(global_settings.data)
			except Exception:
				raise ValidationError('Global Settings should be a correct JSON')

	@classmethod
	def new(cls):
		# update choices
		form = cls()
		form.testcase_sequences.choices = utils.getTestCaseSequences()
		return form


class TestCaseSequenceCreateForm(FlaskForm):
	name = StringField('Name', validators=[DataRequired()])
	description = StringField('Description', validators=[DataRequired()], widget=TextArea())
	classname = SelectField('Class Name')
	datafiles = SelectMultipleField('Data Files')
	testcases = SelectMultipleField('Test Cases')

	@classmethod
	def new(cls, defaults=False):
		# update choices
		form = cls()
		form.datafiles.choices = utils.getDataFiles()
		form.testcases.choices = utils.getTestCases()
		form.classname.choices = utils.getClassNames()

		# set default values
		if defaults:
			form.classname.data = utils.getClassNames(default=utils.CLASSNAME_TESTCASESEQUENCE)

		return form


class TestCaseCreateForm(FlaskForm):
	name = StringField('Name', validators=[DataRequired()])
	description = StringField('Description', validators=[DataRequired()], widget=TextArea())
	classname = SelectField('Class Name')
	browser_type = SelectField('Browser Type')
	testcase_type = SelectField('Test Case Type')
	teststep_sequences = SelectMultipleField('Step Sequences')

	@classmethod
	def new(cls, defaults=False):
		# update choices
		form = cls()
		form.teststep_sequences.choices = utils.getTestStepSequences()
		form.classname.choices = utils.getClassNames()
		form.browser_type.choices = utils.getBrowserTypes()
		form.testcase_type.choices = utils.getTestCaseTypes()

		# set default values
		if defaults:
			form.classname.data = utils.getClassNames(default=utils.CLASSNAME_TESTCASE)
			form.browser_type.data = utils.getBrowserTypes(default=utils.BROWSER_TYPE)
			form.testcase_type.data = utils.getTestCaseTypes(default=utils.TESTCASE_TYPE)

		return form


class TestStepSequenceCreateForm(FlaskForm):
	name = StringField('Name', validators=[DataRequired()])
	description = StringField('Description', validators=[DataRequired()], widget=TextArea())
	classname = SelectField('Class Name')
	teststeps = SelectMultipleField('Test Steps')

	@classmethod
	def new(cls, defaults=False, item_id=None):
		# update choices
		form = cls()
		form.teststeps.choices = utils.getTestSteps(item_id)
		form.classname.choices = utils.getClassNames()

		# set default values
		if defaults:
			form.classname.data = utils.getClassNames(default=utils.CLASSNAME_TESTSTEP)
		
		return form


class TestStepCreateForm(FlaskForm):
	name = StringField('Name', validators=[DataRequired()])
	description = StringField('Description', validators=[DataRequired()], widget=TextArea())
	activity_type = SelectField('Activity Type')
	locator_type = SelectField('Locator Type')
	locator = StringField('Locator')
	value = StringField('Value')
	value2 = StringField('Value 2')
	comparison = SelectField('Comparison')
	optional = SelectField('Optional')
	timeout = StringField('Timeout')
	release = StringField('Release')

	@classmethod
	def new(cls, defaults=False):
		# update choices
		form = cls()
		form.activity_type.choices = utils.getActivityTypes()
		form.locator_type.choices = utils.getLocatorTypes()
		form.comparison.choices = utils.getComparisonChoices()
		form.optional.choices=(('0', 'False'), ('1', 'True'))
		
		# set default values
		if defaults:
			form.activity_type.data = utils.getActivityTypes(default=utils.ACTIVITY_TYPE)
			form.locator_type.data = '0'
			form.comparison.data = '0'
			form.optional.data = '0'

		return form	


	