from flask_login import UserMixin
#from flask import current_app
from werkzeug.security import generate_password_hash, check_password_hash
from datetime import datetime
from app import app, db, login_manager
import uuid
import re
import json
from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION

#
# Handling sclalchemy dialects
#
dialect = re.search(r'(?<=\()\w+', str(db.get_engine()).lower())
# binary fields
if dialect.group() == 'mysql':
	from sqlalchemy.dialects.mysql import BINARY
else:
	BINARY = db.LargeBinary
# double precision
if dialect and dialect.group() == 'postgresql':
	from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION
else:
	DOUBLE_PRECISION = db.Float


#
# UUID as bytes
#
def uuidAsBytes():
	return uuid.uuid4().bytes

#
# user Model
# TODO: extend user Model
#
class User(UserMixin, db.Model):
	__tablename__ = 'users'
	id = db.Column(db.Integer, primary_key=True)
	username = db.Column(db.String(64), unique=True, nullable=False)
	password = db.Column(db.String(128), nullable=False)
	created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
	lastlogin = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)

	def __str__(self):
		return self.username

	def set_password(self, password):
		self.password = generate_password_hash(password)

	def verify_password(self, password):
		return check_password_hash(self.password, password)

@login_manager.user_loader
def load_user(id):
	return User.query.get(int(id))

#
# relation tables
#

testrun_casesequence = db.Table(
	'testrun_casesequence',
	db.Column('testrun_id', BINARY(16), db.ForeignKey('testruns.id'), primary_key=True),
	db.Column('testcase_sequence_id', BINARY(16), db.ForeignKey('testcase_sequences.id'), primary_key=True)
)

testcase_sequence_datafile = db.Table(
	'testcase_sequence_datafile',
	db.Column('testcase_sequence_id', BINARY(16), db.ForeignKey('testcase_sequences.id'), primary_key=True),
	db.Column('datafile_id', BINARY(16), db.ForeignKey('datafiles.id'), primary_key=True)
)

testcase_sequence_case = db.Table(
	'testcase_sequence_case',
	db.Column('testcase_sequence_id', BINARY(16), db.ForeignKey('testcase_sequences.id'), primary_key=True),
	db.Column('testcase_id', BINARY(16), db.ForeignKey('testcases.id'), primary_key=True)
)

testcase_stepsequence = db.Table(
	'testcase_stepsequence',
	db.Column('testcase_id', BINARY(16), db.ForeignKey('testcases.id'), primary_key=True),
	db.Column('teststep_sequence_id', BINARY(16), db.ForeignKey('teststep_sequences.id'), primary_key=True)
)


#
# main entities
#
class Testrun(db.Model):
	#
	# Testrun object
	#

	__tablename__ = 'testruns'
	
	# fields
	id = db.Column(BINARY(16), primary_key=True, default=uuidAsBytes)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)
	created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
	creator_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
	edited = db.Column(db.DateTime, nullable=True)
	editor_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)

	# globals
	global_settings = db.Column(db.String(2048), nullable=True)

	# realshinships
	creator = db.relationship(
		'User',
		backref=db.backref('created_testruns', order_by='desc(Testrun.created)'),
		lazy='select',
		foreign_keys=[creator_id],
	)
	editor = db.relationship('User', backref='edited_testruns', lazy='select', foreign_keys=[editor_id])

	@property
	def uuid(self):
		return str(uuid.UUID(bytes=self.id))

	def __str__(self):
		return self.name

	def __repr__(self):
		return f'<Testrun: {self.uuid}>'

	def finished_calls(self, stage=None):
		if stage == 'None':
			return [call for call in self.calls if (call.is_finished and call.stage == None)]
		if stage:
			return [call for call in self.calls if (call.is_finished and call.stage == stage)]

		return [call for call in self.calls if call.is_finished]


	def to_json(self):
		#
		# returns item as JSON
		#

		return {
			'name': self.name,
			'testrun': {			
				'GC.KWARGS_TESTRUNATTRIBUTES': {
					'GC.EXPORT_FORMAT': {
						'GC.EXPORT_FORMAT': None,
						'GC.EXP_FIELDLIST': [],
					},
					'GC.STRUCTURE_TESTCASESEQUENCE': {
						i+1: self.testcase_sequences[i].to_json(i+1) for i in range(len(self.testcase_sequences))
					},
				}
			},
			'globals' : json.loads(self.global_settings) if self.global_settings else None,
		}

	def get_status(self):
		#
		# check for item completeness
		# returns tuple: (status, message)
		#

		# check for TestCaseSequences
		if len(self.testcase_sequences) == 0:
			return False, f'Testrun {self} does not contain TestCaseSequences'

		# evaluate TestCaseSequences
		for items in self.testcase_sequences:
			status, message = items.get_status()
			if not status:
				return status, message

		return True, 'OK'
		

class TestCaseSequence(db.Model):
	#
	# TestCse Sequence object
	#

	__tablename__ = 'testcase_sequences'

	# fields
	id = db.Column(BINARY(16), primary_key=True, default=uuidAsBytes)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)
	created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
	creator_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
	edited = db.Column(db.DateTime, nullable=True)
	editor_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
	classname_id = db.Column(db.Integer, db.ForeignKey('classnames.id'), nullable=False)
	
	# relashionships
	creator = db.relationship('User', backref='created_testcase_sequences', lazy='select', foreign_keys=[creator_id])
	editor = db.relationship('User', backref='edited_testcase_sequences', lazy='select', foreign_keys=[editor_id])
	classname = db.relationship('ClassName', backref='testcase_sequences', lazy='select', foreign_keys=[classname_id])
	testrun = db.relationship(
		'Testrun',
		secondary=testrun_casesequence,
		lazy='select',
		backref=db.backref('testcase_sequences', order_by='TestCaseSequence.created', lazy=True),
	)

	@property
	def uuid(self):
		return str(uuid.UUID(bytes=self.id))

	def __str__(self):
		return self.name

	def __repr__(self):
		return f'<TestCaseSequence: {self.uuid}>'

	def to_json(self, TestCaseSequenceNumber=1):
		#
		# returns item as JSON
		#

		# get name of DataFile
		if len(self.datafiles) > 0:
			TestDataFileName = self.datafiles[0].uuid
			TestDataSheetName = self.datafiles[0].sheet
		else:
			TestDataFileName = TestDataSheetName = None
		return [
			self.classname.name,
			{
				'SequenceClass': self.classname.name,
				'GC.DATABASE_FILENAME': TestDataFileName,
				'GC.DATABASE_SHEETNAME': TestDataSheetName,
				'GC.EXECUTION_PARALLEL': 1,
				'GC.DATABASE_LINES': '0-999999',
				'GC.STRUCTURE_TESTCASE': {
					i+1: self.testcases[i].to_json(
						TestCaseSequenceNumber,
						i+1,
					) for i in range(len(self.testcases))
				},
			}
		]

	def get_status(self):
		#
		# check for item completeness
		# returns tuple: (status, message)
		#

		# check for DataFiles
		if len(self.datafiles) == 0:
			return False, f'TestCaseSequence {self} does not contain DataFiles'
		
		# check for TestCases
		if len(self.testcases) == 0:
			return False, f'TestCaseSequence {self} does not contain TestCases'

		# evaluate TestCases
		for items in self.testcases:
			status, message = items.get_status()
			if not status:
				return status, message

		return True, 'OK'


class DataFile(db.Model):
	#
	# DataFile object
	#

	__tablename__ = 'datafiles'

	# fields
	id = db.Column(BINARY(16), primary_key=True, default=uuidAsBytes)
	filename = db.Column(db.String(64), nullable=False)
	sheet = db.Column(db.String(32), nullable=False, default='data')
	created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
	creator_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
	
	# relationships
	creator = db.relationship('User', backref='created_datafiles', lazy='select', foreign_keys=[creator_id])
	testcase_sequence = db.relationship(
		'TestCaseSequence',
		secondary=testcase_sequence_datafile,
		lazy='select',
		backref=db.backref('datafiles', lazy=True),
	)

	@property
	def uuid(self):
		return str(uuid.UUID(bytes=self.id))

	@property
	def url(self):
		return '/'.join((
			'http:/',
			app.config.get('BAANGT_DATAFILE_HOST'),
			app.config.get('BAANGT_DATAFILE_GET'),
			self.uuid,
		))
	

	def __str__(self):
		return self.filename

	def __repr__(self):
		return f'<DataFile: {self.uuid}>'


class TestCase(db.Model):
	#
	# testCase object
	#

	__tablename__ = 'testcases'
	
	# fields
	id = db.Column(BINARY(16), primary_key=True, default=uuidAsBytes)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)
	created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
	creator_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
	edited = db.Column(db.DateTime, nullable=True)
	editor_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
	classname_id = db.Column(db.Integer, db.ForeignKey('classnames.id'), nullable=False)
	browser_type_id = db.Column(db.Integer, db.ForeignKey('browser_types.id'), nullable=False)
	testcase_type_id = db.Column(db.Integer, db.ForeignKey('testcase_types.id'), nullable=False)
	
	# realshionshipd
	creator = db.relationship('User', backref='created_testcases', lazy='select', foreign_keys=[creator_id])
	editor = db.relationship('User', backref='edited_testcases', lazy='select', foreign_keys=[editor_id])
	classname = db.relationship('ClassName', backref='testcases', lazy='select', foreign_keys=[classname_id])
	browser_type = db.relationship('BrowserType', backref='testcases', lazy='select', foreign_keys=[browser_type_id])
	testcase_type = db.relationship('TestCaseType', backref='testcases', lazy='select', foreign_keys=[testcase_type_id])
	testcase_sequence = db.relationship(
		'TestCaseSequence',
		secondary=testcase_sequence_case,
		lazy='select',
		backref=db.backref('testcases', order_by='TestCase.created', lazy=True),
	)

	@property
	def uuid(self):
		return str(uuid.UUID(bytes=self.id))

	def __str__(self):
		return self.name

	def __repr__(self):
		return f'<TestCase: {self.uuid}>'

	def to_json(self, TestCaseSequenceNumber=1, TestCaseNumber=1):
		#
		# returns item as JSON
		#

		return [
			self.classname.name,
			{
				'GC.KWARGS_TESTCASETYPE': self.testcase_type.name,
				'GC.KWARGS_BROWSER': self.browser_type.name,
				'GC.BROWSER_ATTRIBUTES': '',
			},
			{
				'GC.STRUCTURE_TESTSTEP': {
					i+1: self.teststep_sequences[i].to_json(
						TestCaseSequenceNumber,
						TestCaseNumber,
						i+1,
					) for i in range(len(self.teststep_sequences))
				},
			}
		]

	def get_status(self):
		#
		# check for item completeness
		# returns tuple: (status, message)
		#

		# check for TestStepSequences
		if len(self.teststep_sequences) == 0:
			return False, f'TestCase {self} does not contain TestStepSequences'

		# evaluate TestStepSequence
		for items in self.teststep_sequences:
			status, message = items.get_status()
			if not status:
				return status, message

		return True, 'OK'


class TestStepSequence(db.Model):
	#
	# TestStep Sequence object
	#

	__tablename__ = 'teststep_sequences'

	# fields
	id = db.Column(BINARY(16), primary_key=True, default=uuidAsBytes)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)
	created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
	creator_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
	edited = db.Column(db.DateTime, nullable=True)
	editor_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
	classname_id = db.Column(db.Integer, db.ForeignKey('classnames.id'), nullable=False)
	
	# relashionship
	creator = db.relationship('User', backref='created_teststep_sequences', lazy='select', foreign_keys=[creator_id])
	editor = db.relationship('User', backref='edited_teststep_sequences', lazy='select', foreign_keys=[editor_id])
	classname = db.relationship('ClassName', backref='teststep_sequences', lazy='select', foreign_keys=[classname_id])
	testcase = db.relationship(
		'TestCase',
		secondary=testcase_stepsequence,
		lazy='select',
		backref=db.backref('teststep_sequences', order_by='TestStepSequence.created', lazy=True),
	)

	@property
	def uuid(self):
		return str(uuid.UUID(bytes=self.id))

	def __str__(self):
		return self.name

	def __repr__(self):
		return f'<TestStepSequence: {self.uuid}>'

	def to_json(self, TestCaseSequenceNumber=1, TestCaseNumber=1, TestStepNumber=1):
		#
		# returns item as JSON
		#

		return [
			{
				'TestStepClass': self.classname.name,
			},
			{
				'GC.STRUCTURE_TESTSTEPEXECUTION': {
					i+1: self.teststeps[i].to_json(
						TestCaseSequenceNumber,
						TestCaseNumber,
						TestStepNumber,
						i+1,
					) for i in range(len(self.teststeps))
				}
			},
		]

	def get_status(self):
		#
		# check for item completeness
		# returns tuple: (status, message)
		#

		# get default classname
		classname = ClassName.query.filter_by(name='GC.CLASSES_TESTSTEPMASTER').first()
		print(f'*** DEFAULT CLASSNAME: {classname.name}')

		# check for TestSteps
		if self.classname == classname and len(self.teststeps) == 0:
			return False, f'TestStepSequence {self} does not contain TestSteps'

		return True, 'OK'


class TestStepExecution(db.Model):
	#
	# TestStep object
	#
	__tablename__ = 'teststep_executions'

	# fields
	id = db.Column(BINARY(16), primary_key=True, default=uuidAsBytes)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)
	created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
	creator_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
	edited = db.Column(db.DateTime, nullable=True)
	locator = db.Column(db.String(512), nullable=True)
	optional = db.Column(db.Boolean, nullable=False, default=False)
	# sa dialect
	timeout = db.Column(DOUBLE_PRECISION(precision='5,2'), nullable=True)
	release = db.Column(db.String(64), nullable=True)
	value = db.Column(db.String(1024), nullable=True)
	value2 = db.Column(db.String(1024), nullable=True)
	comparison = db.Column(db.String(16), nullable=True)
	editor_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
	activity_type_id = db.Column(db.Integer, db.ForeignKey('activity_types.id'), nullable=False)
	locator_type_id = db.Column(db.Integer, db.ForeignKey('locator_types.id'), nullable=True)
	teststep_sequence_id = db.Column(BINARY(16), db.ForeignKey('teststep_sequences.id'), nullable=True)
	
	# relashionships
	creator = db.relationship('User', backref='created_teststeps', lazy='select', foreign_keys=[creator_id])
	editor = db.relationship('User', backref='edited_teststeps', lazy='select', foreign_keys=[editor_id])
	activity_type = db.relationship('ActivityType', backref='teststeps', lazy='select', foreign_keys=[activity_type_id])
	locator_type = db.relationship('LocatorType', backref='teststeps', lazy='select', foreign_keys=[locator_type_id])
	teststep_sequence = db.relationship(
		'TestStepSequence',
		backref=db.backref('teststeps', order_by='TestStepExecution.created'),
		lazy='select',
		foreign_keys=[teststep_sequence_id],
	)
	
	@property
	def uuid(self):
		return str(uuid.UUID(bytes=self.id))

	def __str__(self):
		return self.name

	def __repr__(self):
		return f'<TestStepExecution: {self.uuid}>'

	def to_json(self, TestCaseSequenceNumber=1, TestCaseNumber=1, TestStepNumber=1, TestStepExecutionNumber=1):
		#
		# returns item as JSON
		#

		# set LocatorType name
		#locator_type = ''
		#if self.locator_type:
		#	locator_type = self.locator_type.name
			
		return {
			'Activity': self.activity_type.name,
			'LocatorType': self.locator_type.name if self.locator_type else '',
			'Locator': self.locator or '',
			'Value': self.value or '',
			'Comparison': self.comparison or '',
			'Value2': self.value2 or '',
			'Timeout': self.timeout or '',
			'Optional': self.optional or '',
			'Release': self.release or '',
		}

#
# supporting entities
#
class GlobalTestStepExecution(db.Model):
	__tablename__ = 'global_teststep_executions'
	id = db.Column(BINARY(16), primary_key=True, default=uuidAsBytes)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)
	created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
	creator_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
	edited = db.Column(db.DateTime, nullable=True)
	editor_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
	activity_type_id = db.Column(db.Integer, db.ForeignKey('activity_types.id'), nullable=False)
	locator_type_id = db.Column(db.Integer, db.ForeignKey('locator_types.id'), nullable=False)
	teststep_sequence_id = db.Column(BINARY(16), db.ForeignKey('teststep_sequences.id'), nullable=True)

	# relationships
	creator = db.relationship('User', backref='created_global_teststeps', lazy='select', foreign_keys=[creator_id])
	editor = db.relationship('User', backref='edited_global_teststeps', lazy='select', foreign_keys=[editor_id])
	activity_type = db.relationship('ActivityType', backref='global_teststeps', lazy='select', foreign_keys=[activity_type_id])
	locator_type = db.relationship('LocatorType', backref='global_teststeps', lazy='select', foreign_keys=[locator_type_id])
	teststep_sequence = db.relationship('TestStepSequence', backref='global_teststeps', lazy='select', foreign_keys=[teststep_sequence_id])
	
	@property
	def uuid(self):
		return str(uuid.UUID(bytes=self.id))

	def __str__(self):
		return self.name


class ClassName(db.Model):
	__tablename__ = 'classnames'
	id = db.Column(db.Integer, primary_key=True)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)

	def __str__(self):
		return self.name

class BrowserType(db.Model):
	__tablename__ = 'browser_types'
	id = db.Column(db.Integer, primary_key=True)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)

	def __str__(self):
		return self.name
	
class TestCaseType(db.Model):
	__tablename__ = 'testcase_types'
	id = db.Column(db.Integer, primary_key=True)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)

	def __str__(self):
		return self.name

class ActivityType(db.Model):
	__tablename__ = 'activity_types'
	id = db.Column(db.Integer, primary_key=True)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)

	def __str__(self):
		return self.name

class LocatorType(db.Model):
	__tablename__ = 'locator_types'
	id = db.Column(db.Integer, primary_key=True)
	name = db.Column(db.String(64), nullable=False)
	description = db.Column(db.String(512), nullable=False)

	def __str__(self):
		return self.name


#
# Testrun results
#

class TestRunCall(db.Model):
	__tablename__ = 'testrun_calls'
	id = db.Column(BINARY(16), primary_key=True)
	created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
	creator_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
	testrun_id = db.Column(BINARY(16), db.ForeignKey('testruns.id'), nullable=False)

	# relationships
	creator = db.relationship('User', backref='testruns_called', lazy='select', foreign_keys=[creator_id])
	testrun = db.relationship(
		'Testrun',
		backref=db.backref('calls', order_by='TestRunCall.created'),
		lazy='select',
		foreign_keys=[testrun_id],
	)

	# Testrun call faild
	is_failed = db.Column(db.Boolean, nullable=False, default=False)
	error_message = db.Column(db.String(512), nullable=True) 

	# summary
	is_finished = db.Column(db.Boolean, nullable=False, default=False)
	testrecords = db.Column(db.Integer, nullable=True)
	successful = db.Column(db.Integer, nullable=True)
	error = db.Column(db.Integer, nullable=True)
	paused = db.Column(db.Integer, nullable=True)
	duration = db.Column(db.Integer, nullable=True)
	stage = db.Column(db.String(32), nullable=True)

	def __str__(self):
		return str(uuid.UUID(bytes=self.id))

