import os
import re
import xlrd
import json
from app import models, db
from app.utils import idAsBytes
from app.utils.datafile_handler import uploadDataFile
from flask import current_app
from flask_login import current_user
from sqlalchemy import func
from datetime import datetime


class TestrunXLSXParser:
    #
    # parses XLSX definition to TestrunNodes
    #
    
    # required sheets and fields of complex format
    COMPLEX_KEYS = {
        'TestCaseSequence': [
            (
                'Number',
            ),
            {
                'field': 'SequenceClass',
                'attr_name': 'classname',
            },
            {
                'field': 'TestDataFileName',
                'attr_name': 'datafile_name',
            },
            {
                'field': 'Sheetname',
                'attr_name': 'datafile_sheet',
            },
        ],
        'TestCase': [
            (
                'TestCaseSequenceNumber',
                'TestCaseNumber',
            ),
            {
                'field': 'TestCaseClass',
                'attr_name': 'classname',
            },
            {
                'field': 'TestCaseType',
                'attr_name': 'testcase_type',
            },
            {
                'field': 'Browser',
                'attr_name': 'browser_type',
            },
        ],
        'TestStep': [
            (
                'TestCaseSequenceNumber',
                'TestCaseNumber',
                'TestStepNumber',
            ),
            {
                'field': 'TestStepClass',
                'attr_name': 'classname',
            },
        ],
        'TestStepExecution': [
            (
                'TestCaseSequenceNumber',
                'TestCaseNumber',
                'TestStepNumber',
                'TestStepExecutionNumber',
            ),
            {
                'field': 'Activity',
                'attr_name': 'activity_type',
            },
            {
                'field': 'LocatorType',
                'attr_name': 'locator_type',
            },
            {
                'field': 'Locator',
                'attr_name': 'locator',
            },
            {
                'field': 'Value',
                'attr_name': 'value',
            },
            {
                'field': 'Comparison',
                'attr_name': 'comparison',
            },
            {
                'field': 'Value2',
                'attr_name': 'value2',
            },
            {
                'field': 'Timeout',
                'attr_name': 'timeout',
            },
            {
                'field': 'Optional',
                'attr_name': 'optional',
            },
            {
                'field': 'Release',
                'attr_name': 'release',
            },
        ],
    }

    # required sheets and fields of simple format
    SIMPLE_KEYS = {
        'TestStepExecution': {
            'Activity': 'activity_type',
            'LocatorType': 'locator_type',
            'Locator': 'locator',
            'Value': 'value',
            'Comparison': 'comparison',
            'Value2': 'value2',
            'Timeout': 'timeout',
            'Optional': 'optional',
            'Release': 'release',
        },
    }

    def __init__(self, content):
        # filename
        self.filename = os.path.basename(content.filename)
        # get workbook
        self.workbook = xlrd.open_workbook(file_contents=content.read())
        # check definition format
        self.is_simple = not all(sheet in self.workbook.sheet_names() for sheet in self.COMPLEX_KEYS)
        self.assert_workbook()

    def assert_workbook(self):
        sheet_name = next(iter(self.SIMPLE_KEYS))
        if self.is_simple:
            if not sheet_name in self.workbook.sheet_names():
                raise Exception(f'Invalid definition format: sheet "{sheet_name}" is absent')
            if not all(field in self.workbook.sheet_by_name(sheet_name).row_values(0) for field in self.SIMPLE_KEYS[sheet_name]):
                raise Exception(
                    f'Invalid definition format: sheet "{sheet_name}" defines not all required fields {self.SIMPLE_KEYS[sheet_name]}'
                )
        else:
            for sheet, fields in self.COMPLEX_KEYS.items():
                required_fields = [i.get('field') for i in fields[1:]]
                required_fields += list(fields[0])
                if not all(field in self.workbook.sheet_by_name(sheet).row_values(0) for field in required_fields):
                    raise Exception(f'Invalid definition format: sheet "{sheet}" defines not all required fields {fields}')

    def simple_nodes(self):
        #
        # generator
        # parses simple definition file
        # yields TestStepExecution nodes
        #

        # get TestStepExecution sheet
        sheet_name = next(iter(self.SIMPLE_KEYS))
        sheet = self.workbook.sheet_by_name(sheet_name)

        # parse TestSteps
        sheet_headers = {h: i for i, h in enumerate(sheet.row_values(0))}
        for row in range(1, sheet.nrows):
            # create new TestStep Node
            teststep = TestrunNode(TestrunNode.TSE, row)
            # set attributes
            for field, attr_name in self.SIMPLE_KEYS[sheet_name].items():
                teststep.set_attr(attr_name, sheet.cell(row, sheet_headers[field]).value or None)

            yield teststep

    def simple_datafile(self):
        #
        # returns tuple
        # (datafile name, data sheet)
        # for simple definition
        #

        if 'data' in self.workbook.sheet_names():
            return self.filename, 'data'


    def complex_nodes(self):
        #
        # generator
        # parses complex definition file
        # yields tuple of 2 elements:
        # [0]: parent id 
        # [1]: child node
        #

        # parse sheets
        for sheet_name, headers in self.COMPLEX_KEYS.items():
            # get sheet
            sheet = self.workbook.sheet_by_name(sheet_name)
            # parse sheet
            sheet_headers = {h: i for i, h in enumerate(sheet.row_values(0))}
            for row in range(1, sheet.nrows):
                # get parent nodes numbers
                parent_numbers = list(map(lambda col: int(sheet.cell(row, sheet_headers[col]).value), headers[0][:-1]))
                
                # get child number
                try:
                    child_number = int(sheet.cell(row, sheet_headers[headers[0][-1]]).value)
                except Exception:
                    child_number = row

                # create new child
                child = TestrunNode(sheet_name, child_number)
                # set child's attributes
                for attr in headers[1:]:
                    child.set_attr(
                        attr.get('attr_name'),
                        sheet.cell(row, sheet_headers[attr.get('field')]).value or None)

                yield parent_numbers, child



#
# Testrun Tree
#
class TestrunNode:
    #
    # Testrun Tree Node
    # (represents single Testrun layer instance)
    #

    # layers
    TR = 0  # Testrun
    TCS = 1 # TestCaseSequence
    TC = 2  # TestCase
    TS = 3 # TestStep
    TSE = 4  # TestStepExecution 
    
    # node labels
    LABELS = {
        'Testrun': TR,
        'TestCaseSequence': TCS,
        'TestCase': TC,
        'TestStep': TS,
        'TestStepExecution': TSE,
    }

    def __init__(self, layer, node_number):
        self.set_layer(layer)
        self.set_number(node_number)
        self.attrs = {}
        self.children = []

    def __str__(self):
        return f'{self.label}-{self.number}: {", ".join(map(str, self.attrs.values()))}'

    @property
    def label(self):
        #
        # label of the node (see LABELS)
        #
        return {value: key for key, value in self.LABELS.items()}.get(self.layer)

    def add_child(self, child):
        #
        # inserts child node according to its number
        #
        self.children.append(child)

    def update_children(self):
        #
        # updates children nodes:
        # - removes TestCaseSequence, TestCase, TestStep nodes without classnames
        # - sorts children nodes by their number
        # - updates node numbers to be sequentialy from 1
        #

        # remove children nodes without classname
        if self.layer in [self.TR, self.TCS, self.TC]:
            children = filter(lambda child: bool(child.get_attr('classname')), self.children)
        else:
            children = (child for child in self.children)

        # sort children
        self.children = sorted(children, key=lambda child: child.number)

        # update child numbers
        for index, child in enumerate(self.children, 1):
            child.set_number(index)

    def set_number(self, number):
        #
        # sets number property
        #
        self.number = number

    def set_layer(self, layer):
        #
        # sets layer by integer or label
        #
        if type(layer) == type(1):
            self.layer = layer
        else:
            self.layer = self.LABELS.get(layer)

    def set_datafile(self, data):
        #
        # sets datafile and sheet attributes of TCS node
        #
        if self.layer != self.TCS:
            return False

        self.set_attr('datafile_name', data[0])
        self.set_attr('datafile_sheet', data[1])
        return True

    def set_attr(self, key, value):
        #
        # sets attribute to {key: value}
        #
        self.attrs[key] = value

    def get_attr(self, key):
        #
        # returns key atrribute
        #
        return self.attrs.get(key)

    def  get_child(self, number):
        for child in self.children:
            if child.number == number:
                return child

    def get_node(self, numbers):
        #
        # returns node from tree by node numbers
        #

        if not numbers:
            return self

        node = self.get_child(numbers[0])
        if node:
            # go deeper into children
            return node.get_node(numbers[1:])

    def pprint(self, indent='|'):
        print(f'{indent}{self}')
        for child in self.children:
            child.pprint(indent=f'{indent}-')
        


class TestrunTree:
    #
    # Testrun Tree 
    #

    # default attributes
    DEFAULT_CLASSNAME_TESTCASESEQUENCE = 'GC.CLASSES_TESTCASESEQUENCE'
    DEFAULT_CLASSNAME_TESTCASE = 'GC.CLASSES_TESTCASE'
    DEFAULT_CLASSNAME_TESTSTEP = 'GC.CLASSES_TESTSTEPMASTER'
    DEFAULT_BROWSER_TYPE = 'FF'
    DEFAULT_TESTCASE_TYPE = 'Browser'

    JSON_KEYS = {
        'root': 'GC.KWARGS_TESTRUNATTRIBUTES',
        'TestCaseSequence': {
            'structure': 'GC.STRUCTURE_TESTCASESEQUENCE',
            'datafile_name': 'GC.DATABASE_FILENAME',
            'datafile_sheet': 'GC.DATABASE_SHEETNAME',
        },
        'TestCase': {
            'structure': 'GC.STRUCTURE_TESTCASE',
            'testcase_type': 'GC.KWARGS_TESTCASETYPE',
            'browser_type': 'GC.KWARGS_BROWSER',
        },
        'TestStep': {
            'structure': 'GC.STRUCTURE_TESTSTEP',
            'classname': 'TestStepClass',
        },
        'TestStepExecution': {
            'structure': 'GC.STRUCTURE_TESTSTEPEXECUTION',
            'Activity': 'activity_type',
            'LocatorType': 'locator_type',
            'Locator': 'locator',
            'Value': 'value',
            'Comparison': 'comparison',
            'Value2': 'value2',
            'Timeout': 'timeout',
            'Optional': 'optional',
            'Release': 'release',
        },
    }
        

    def __init__(self, simple=False):
        if simple:
            # create default tree
            child_node = None
            for layer in reversed(range(TestrunNode.TSE)):
                node = TestrunNode(layer, 1)
                if child_node:
                    node.add_child(child_node)
                child_node = node
            self.root = node
            self.set_defaults()
        else:
            self.root = TestrunNode(TestrunNode.TR, 1)

    @classmethod
    def default_classnames(cls):
        return {
            cls.DEFAULT_CLASSNAME_TESTCASESEQUENCE,
            cls.DEFAULT_CLASSNAME_TESTCASE,
            cls.DEFAULT_CLASSNAME_TESTSTEP,
        }
    
    @classmethod
    def from_xlsx(cls, content):
        #
        # Builds TestrunTree from XLSX content
        #

        parser = TestrunXLSXParser(content)
        if parser.is_simple:
            tree = cls(simple=True)
            # set datafile
            tcs = tree.get_node((1,)) # get TestCaseSequence
            tcs.set_datafile(parser.simple_datafile())
            # set steps
            tss = tree.get_node((1, 1, 1)) # get TestStepSequence 
            for node in parser.simple_nodes():
                tss.add_child(node)
        else:
            tree = cls()
            for numbers, child in parser.complex_nodes():
                parent = tree.get_node(numbers)
                parent.add_child(child)
            tree.update()

        return tree

    @classmethod
    def from_json(cls, content):
        #
        # Builds TestrunTree from JSON content
        #

        # load JSON definition
        testrun = json.load(content).get(cls.JSON_KEYS['root'])
        tree = cls()
        
        # parse TestCaseSequences
        for tcs_number, tcs_data in testrun.get(cls.JSON_KEYS['TestCaseSequence'].get('structure')).items():
            # create TestCaseSequences node
            tcs = TestrunNode('TestCaseSequence', int(tcs_number))
            # set attributes
            tcs.set_attr('classname', tcs_data[0])
            tcs.set_attr('datafile_name', tcs_data[1].get(cls.JSON_KEYS['TestCaseSequence'].get('datafile_name')))
            tcs.set_attr('datafile_sheet', tcs_data[1].get(cls.JSON_KEYS['TestCaseSequence'].get('datafile_sheet')))
            # apend to parent Testrun
            tree.root.add_child(tcs)

            # parse TestCases
            for tc_number, tc_data in tcs_data[1].get(cls.JSON_KEYS['TestCase'].get('structure')).items():
                # create TestCase node
                tc = TestrunNode('TestCase', int(tc_number))
                # set attributes
                tc.set_attr('classname', tc_data[0])
                tc.set_attr('testcase_type', tc_data[1].get(cls.JSON_KEYS['TestCase'].get('testcase_type')))
                tc.set_attr('browser_type', tc_data[1].get(cls.JSON_KEYS['TestCase'].get('browser_type')))
                # append to parent TestCaseSequence
                tcs.add_child(tc)

                # parse TestSteps
                for ts_number, ts_data in tc_data[2].get(cls.JSON_KEYS['TestStep'].get('structure')).items():
                    # create TestStep node
                    ts = TestrunNode('TestStep', int(ts_number))
                    # set attributes
                    ts.set_attr('classname', ts_data[0].get(cls.JSON_KEYS['TestStep'].get('classname')))
                    # append to parent TestCase
                    tc.add_child(ts)

                    # check if TestStepExecutions are defined
                    if len(ts_data) > 1 and ts_data[1].get(cls.JSON_KEYS['TestStepExecution'].get('structure')):
                        # parse TestStepExeutions
                        for tse_number, tse_data in ts_data[1].get(cls.JSON_KEYS['TestStepExecution'].get('structure')).items():
                            # create TestStepExecution node
                            tse = TestrunNode('TestStepExecution', int(tse_number))
                            # set attributes
                            for field, attr_name in cls.JSON_KEYS['TestStepExecution'].items():
                                field_value = tse_data.get(field)
                                if field_value:
                                    tse.set_attr(attr_name, field_value)
                            # append to parent TestStep
                            ts.add_child(tse)

        tree.update()
        return tree


    def update(self, node=None):
        #
        # updates TestrunTree:
        # - removes TestCaseSequence, TestCase, TestStep nodes without classnames
        # - sorts children nodes by their number
        # - updates node numbers to be sequentialy from 1
        #

        if node is None:
            node = self.root

        node.update_children()
        # go deeper
        for child in node.children:
            self.update(node=child)

    def set_defaults(self):
        #
        # sets default attributes to layer nodes
        #

        # TestCaseSequence
        tcs = self.get_node((1,))
        tcs.set_attr('classname', self.DEFAULT_CLASSNAME_TESTCASESEQUENCE)
        
        # TestCase
        tc = self.get_node((1, 1))
        tc.set_attr('classname', self.DEFAULT_CLASSNAME_TESTCASE)
        tc.set_attr('testcase_type', self.DEFAULT_TESTCASE_TYPE)
        tc.set_attr('browser_type', self.DEFAULT_BROWSER_TYPE)
        
        # TestStepSequence
        tss = self.get_node((1, 1, 1))
        tss.set_attr('classname', self.DEFAULT_CLASSNAME_TESTSTEP) 

    def get_node(self, node_numbers):
        return self.root.get_node(node_numbers)


#
# import Testrun from XLSX or JSON definition
#
def importTestrun(content, datafiles=None, testrun_id=None):
    #
    # imports testrun from xlsx or json file
    #

    # get Testrun name
    testrun_name = os.path.basename(content.filename)

    # check for import option: create or update
    if testrun_id:
        # get item
        testrun = models.Testrun.query.get(idAsBytes(testrun_id))
        if testrun is None: 
            # item does not exist
            raise Exception(f'Testrun {testrun_id} does not exists.')
    else:
        # check if imported Testrun already exists
        testrun =  models.Testrun.query.filter_by(name=testrun_name).first()
        if testrun:
            testrun_id = testrun.uuid

    # parse testrun content
    source_type = testrun_name.split('.')[-1].upper()
    if source_type == 'XLSX':
        testrun_tree = TestrunTree.from_xlsx(content)
        current_app.logger.info(f'Testrun successfully read from {content.filename} by {current_user}.')
    elif source_type == 'JSON':
        testrun_tree = TestrunTree.from_json(content)
        current_app.logger.info(f'Testrun successfully read from {content.filename} by {current_user}.')
        print('*** TREE:')
        testrun_tree.root.pprint()
    else:
        raise Exception(f'File type {source_type} is not supported')
    
    if testrun:
        # update mode
        current_app.logger.info(f'Updating Testrun {testrun.uuid} from {content.filename} by {current_user}.')

        testrun.description = f'Updated from "{testrun_name}"'
        testrun.editor = current_user
        testrun.edited = datetime.utcnow()
    else:
        # create mode
        current_app.logger.info(f'Importing Testrun from {content.filename} by {current_user}.')

        # create Testrun
        testrun  = models.Testrun(
            name=testrun_name,
            description=f'Imported from "{testrun_name}"',
            creator=current_user,
        )
        db.session.add(testrun)

    # TestCaseSequences
    for tcs_index, tcs in enumerate(testrun_tree.root.children):
        # ClassName
        classname = getOrCreateClassNameByName(tcs.get_attr('classname'), testrun_name)
        # DataFile
        datafile_name = tcs.get_attr('datafile_name')
        datafile_sheet = tcs.get_attr('datafile_sheet')
        # get datafile
        try:
            datafile = testrun.testcase_sequences[tcs_index].datafiles[0]
        except Exception:
            datafile = None
        # upload datafile
        datafile_id = None
        if datafile_name == testrun_name:
            content.seek(0)
            datafile_id = uploadDataFile(content, datafile.uuid if datafile else None)
        else:
            try:
                datafile_name = os.path.basename(datafile_name)
                new_datafile = next(filter(
                    lambda df: df.data and os.path.basename(df.data.filename) == datafile_name,
                    datafiles,
                )).data
                #new_datafile = next(filter(lambda df: df.data and os.path.basename(df.data.filename) == datafile_name, datafiles)).data
            except Exception:
                new_datafile = None
            if new_datafile:
                datafile_id = uploadDataFile(new_datafile, datafile.uuid if datafile else None)
        if datafile_id and datafile:
            datafile.filename = datafile_name
            datafile.sheet = datafile_sheet
        elif datafile_id:
            datafile = models.DataFile(
                id=idAsBytes(datafile_id),
                filename=datafile_name,
                sheet=datafile_sheet,
                creator=current_user,
            )
            db.session.add(datafile)
        
        if tcs_index < len(testrun.testcase_sequences):
            # update mode
            testcase_sequence = testrun.testcase_sequences[tcs_index]
            testcase_sequence.description = f'Updated from "{testrun_name}"'
            testcase_sequence.editor = current_user
            testcase_sequence.edited = datetime.utcnow()
            testcase_sequence.classname = classname
            testcase_sequence.datafiles = [datafile] if datafile else []
        else:
            # create mode
            testcase_sequence = models.TestCaseSequence(
                name=f'{testrun_name}_{tcs.number}',
                description=f'Imported from "{testrun_name}"',
                creator=current_user,
                classname=classname,
                datafiles=[datafile] if datafile else [],
                testrun=[testrun],
            )
            db.session.add(testcase_sequence)

        # TestCases
        for tc_index, tc in enumerate(tcs.children):
            # ClassName
            classname = getOrCreateClassNameByName(tc.get_attr('classname'), testrun_name)
            # TestCaseType
            testcase_type=getTestCaseTypeByName(tc.get_attr('testcase_type'))
            # BrowserType
            browser_type = getBrowserTypeByName(tc.get_attr('browser_type'))
            if tc_index < len(testcase_sequence.testcases):
                # update mode
                testcase = testcase_sequence.testcases[tc_index]
                testcase.description = f'Updated from "{testrun_name}"'
                testcase.editor = current_user
                testcase.edited = datetime.utcnow()
                testcase.classname = classname
                testcase.testcase_type = testcase_type
                testcase.browser_type = browser_type
            else:
                # create mode
                testcase  = models.TestCase(
                    name=f'{testrun_name}_{tcs.number}_{tc.number}',
                    description=f'Imported from "{testrun_name}"',
                    creator=current_user,
                    classname=classname,
                    browser_type=browser_type,
                    testcase_type=testcase_type,
                    testcase_sequence=[testcase_sequence],
                )
                db.session.add(testcase)

            # TestStepSequences
            for tss_index, tss in enumerate(tc.children):
                # ClassName
                classname = getOrCreateClassNameByName(tss.get_attr('classname'), testrun_name)
                if  tss_index < len(testcase.teststep_sequences):
                    # update mode
                    teststep_sequence = testcase.teststep_sequences[tss_index]
                    teststep_sequence.description = f'Updated from "{testrun_name}"'
                    teststep_sequence.editor = current_user
                    teststep_sequence.edited = datetime.utcnow()
                    teststep_sequence.classname = classname
                else:
                    # create mode
                    teststep_sequence  = models.TestStepSequence(
                        name=f'{testrun_name}_{tcs.number}_{tc.number}_{tss.number}',
                        description=f'Imported from "{testrun_name}"',
                        creator=current_user,
                        classname=classname,
                        testcase=[testcase],
                    )
                    db.session.add(teststep_sequence)

                # TestStep
                for ts_index, ts in enumerate(tss.children):
                    # Activity Type
                    activity_type = getOrCreateActivityTypeByName(ts.get_attr('activity_type'), testrun_name)
                    # TestCaseType
                    locator_type=getLocatorTypeByName(ts.get_attr('locator_type'))
                    # value
                    int_parser = lambda value: int(value) if re.match(r'\d+\.0+$', str(value)) else value
                    value = int_parser(ts.get_attr('value'))
                    # value2
                    value2 = int_parser(ts.get_attr('value2'))
                    # optional
                    optional = bool(ts.get_attr('optional'))
                    if ts_index < len(teststep_sequence.teststeps):
                        # update mode
                        teststep = teststep_sequence.teststeps[ts_index]
                        teststep.description = f'Updated from "{testrun_name}"'
                        teststep.editor = current_user
                        teststep.edited = datetime.utcnow()
                        teststep.activity_type = activity_type
                        teststep.locator_type = locator_type
                        teststep.locator = ts.get_attr('locator')
                        teststep.value = value
                        teststep.comparison = ts.get_attr('comparison')
                        teststep.value2 = value2
                        teststep.timeout = ts.get_attr('timeout')
                        teststep.optional = optional
                        teststep.release = ts.get_attr('release')
                    else:
                        # create mode
                        teststep  = models.TestStepExecution(
                            name=f'{testrun_name}_{tcs.number}_{tc.number}_{tss.number}_{ts.number}',
                            description=f'Imported from "{testrun_name}"',
                            creator=current_user,
                            teststep_sequence=teststep_sequence,
                            activity_type=activity_type,
                            locator_type=locator_type,
                            locator=ts.get_attr('locator'),
                            value=value,
                            comparison=ts.get_attr('comparison'),
                            value2=value2,
                            timeout=ts.get_attr('timeout'),
                            optional=optional,
                            release=ts.get_attr('release'),
                        )
                        db.session.add(teststep)
                # remove old extra items
                teststep_sequence.teststeps = teststep_sequence.teststeps[:len(tss.children)]
            testcase.teststep_sequences = testcase.teststep_sequences[:len(tc.children)]
        testcase_sequence.testcases = testcase_sequence.testcases[:len(tcs.children)]
    testrun.testcase_sequences = testrun.testcase_sequences[:len(testrun_tree.root.children)]

    db.session.commit()


#
# support instances getters
#
def getOrCreateClassNameByName(name, source_filename):
    # check if name not empty
    if not name or not name.strip():
        return None

    # get ClassName from DB
    if not name in TestrunTree.default_classnames():
        name = f'plugin.{name}'
    classname = models.ClassName.query.filter_by(name=name).first()
    if classname is None:
        # create ClassName if it doesn't exist
        classname = models.ClassName(
            name=name,
            description=f'Imported from {source_filename}',
        )
        db.session.add(classname)
        #db.session.commit()
        #current_app.logger.info(f'Created ClassName ID #{classname.id} by {current_user}.')
        current_app.logger.info(f'Added ClassName  "{classname.name}" by {current_user}.')

    return classname


def getBrowserTypeByName(name):
    # browser mapper
    bm = {
        'BROWSER_FIREFOX': "FF",
        'BROWSER_CHROME': "Chrome",
        'BROWSER_SAFARI': "Safari",
        'BROWSER_EDGE': "Edge",
    }
    if bm.get(name.split('.')[-1]):
        return models.BrowserType.query.filter_by(name=bm[name.split('.')[-1]]).first()
    return models.BrowserType.query.filter_by(name=name).first()

def getTestCaseTypeByName(name):
    return models.TestCaseType.query.filter_by(name=name).first()

def getOrCreateActivityTypeByName(name, source_filename):
    # check if name not empty
    if not name or not name.strip():
        return None

    # get ActivityType from DB
    activity = models.ActivityType.query.filter(func.upper(models.ActivityType.name) == name.upper()).first()
    if activity is None:
        # create Activity if it doesn't exist
        activity = models.ActivityType(
            name=name,
            description=f'Imported from {source_filename}',
        )
        db.session.add(activity)
        #db.session.commit()
        #current_app.logger.info(f'Created ClassName ID #{classname.id} by {current_user}.')
        current_app.logger.info(f'Added ActivityType  "{activity.name}" by {current_user}.')

    return activity 

def getLocatorTypeByName(name):
    if name:
        #return models.LocatorType.query.filter(func.upper(models.LocatorType.name) == name.upper()).first()
        return models.LocatorType.query.filter_by(name=name).first()
    else:
        return None






