
from flask import render_template, redirect, flash, request, url_for, send_from_directory, jsonify, abort
from flask_login import login_required, current_user, login_user, logout_user
from flask.logging import default_handler
from app import app, db, models, forms, utils
#from app.utils.xlsx_handler import importXLSX
from app.utils.testrun_import import importTestrun
from app.charts import DashboardCharts, SummaryCharts
from datetime import datetime
from sqlalchemy import func, desc
import requests
import json
import uuid
import os

# handle favicon requests
@app.route('/favicon.ico') 
def favicon(): 
	return send_from_directory('static/media', 'favicon.ico', mimetype='image/vnd.microsoft.icon')

@app.route('/')
@login_required
def index():
	return render_template('testrun/index.html', items=utils.getItemCategories())

@app.route('/<string:item_type>')
@login_required
def item_list(item_type):
	# placeholder for import form
	form = None
	# get item list by type
	if item_type == 'testrun':
		items = models.Testrun.query.order_by(desc('created')).all()
		# build form for importing a testrun
		form = forms.TestrunImportForm()
		
	elif item_type == 'testcase_sequence':
		items = models.TestCaseSequence.query.order_by(desc('created')).all()
		# DataFile update form
		form = forms.DataFileUpdateForm()
	elif item_type == 'testcase':
		items = models.TestCase.query.order_by(desc('created')).all()
	elif item_type == 'teststep_sequence':
		items = models.TestStepSequence.query.order_by(desc('created')).all()
	elif item_type == 'teststep':
		items = models.TestStepExecution.query.order_by(desc('created')).all()
	else:
		app.logger.warning(f'Item type "{item_type}" does not exist. Requested by "{current_user}".')
		flash(f'Item type "{item_type}" does not exist.', 'warning')
		return redirect(url_for('index'))

	return render_template('testrun/item_list.html', type=item_type, items=items, form=form)


@app.route('/<string:item_type>/<string:item_id>/delete', methods=['POST', 'DELETE'])
@login_required
def delete_item(item_type, item_id):
	#
	# delete item
	#

	# cascade delete
	if request.method == 'POST':
		app.logger.info(f'Cascade deletion of {utils.getItemType(item_type)} {item_id} requested by "{current_user}".')
		try:
			utils.deleteCascade(item_type, uuid.UUID(item_id).bytes)
			flash(f'Item {utils.getItemType(item_type)} {item_id} and its children have been successfully deleted.', 'success')
			app.logger.info(
				f'Item {utils.getItemType(item_type)} {item_id} and its children have been successfully deleted by "{current_user}".'
			)
			return jsonify({'success': 'OK'}), 200
		except Exception as error:
			app.logger.error(f'Failed to cascade delete {utils.getItemType(item_type)} {item_id} requested by "{current_user}": {error}')
			return jsonify({'error': 'Bad Request'}), 400

	# delete single item
	# DELETE method
	app.logger.info(f'Deletion of {utils.getItemType(item_type)} {item_id} requested by "{current_user}".')
	#elif request.method == 'DELETE':
	
	# get item by type and id	
	if item_type == 'testrun':
		item = models.Testrun.query.get(utils.idAsBytes(item_id))
	elif item_type == 'testcase_sequence':
		item = models.TestCaseSequence.query.get(utils.idAsBytes(item_id))
	elif item_type == 'testcase':
		item = models.TestCase.query.get(utils.idAsBytes(item_id))
	elif item_type == 'teststep_sequence':
		item = models.TestStepSequence.query.get(utils.idAsBytes(item_id))
	elif item_type == 'teststep':
		item = models.TestStepExecution.query.get(utils.idAsBytes(item_id))
	else:
		app.logger.warning(f'Item type "{item_type}" does not exist. Requested by "{current_user}".')
		flash(f'Item type "{item_type}" does not exist.', 'warning')
		return jsonify({'error': 'Bad Request'}), 400

	if item is None:
		app.logger.error(f'Item {utils.getItemType(item_type)} {item_id} does not exist.')
		return jsonify({'error': 'Bad Request'}), 400

	db.session.delete(item)
	db.session.commit()
	app.logger.info(f'Item {utils.getItemType(item_type)} {item_id} successfully deleted by {current_user}.')
	flash(f'Item "{item.name}" has been successfully deleted.', 'success')
	return jsonify({'success': 'OK'}), 200


@app.route('/<string:item_type>/<string:item_id>/edit', methods=['GET', 'POST'])
@login_required
def edit_item(item_type, item_id):
	#
	# edit item
	#

	idAsBytes = uuid.UUID(item_id).bytes
	if item_type == 'testrun':
		item = models.Testrun.query.get(idAsBytes)
		form = forms.TestrunCreateForm.new()
		if request.method == 'GET':
			form.global_settings.data = item.global_settings
			form.testcase_sequences.data = [f'{x.uuid}' for x in item.testcase_sequences]
	elif item_type == 'testcase_sequence':
		item = models.TestCaseSequence.query.get(idAsBytes)
		form = forms.TestCaseSequenceCreateForm.new()
		if request.method == 'GET':
			form.classname.data = f'{item.classname.id}'
			form.datafiles.data = [f'{x.uuid}' for x in item.datafiles]
			form.testcases.data = [f'{x.uuid}' for x in item.testcases]
	elif item_type == 'testcase':
		item = models.TestCase.query.get(idAsBytes)
		form = forms.TestCaseCreateForm.new()
		if request.method == 'GET':
			form.classname.data = f'{item.classname.id}'
			form.browser_type.data = f'{item.browser_type.id}'
			form.testcase_type.data = f'{item.testcase_type.id}'
			form.teststep_sequences.data = [f'{x.uuid}' for x in item.teststep_sequences]
	elif item_type == 'teststep_sequence':
		item = models.TestStepSequence.query.get(idAsBytes)
		form = forms.TestStepSequenceCreateForm.new(item_id=item_id)
		if request.method == 'GET':
			form.classname.data = f'{item.classname.id}'
			form.teststeps.data =[f'{x.uuid}' for x in item.teststeps]
	elif item_type == 'teststep':
		item = models.TestStepExecution.query.get(idAsBytes)
		form = forms.TestStepCreateForm.new()
		if request.method == 'GET':
			form.activity_type.data = f'{item.activity_type.id}'
			form.locator_type.data = f'{item.locator_type.id}' if item.locator_type else '0'
			# model extension
			form.locator.data = item.locator or ''
			if item.optional:
				form.optional.data = '1'
			else:
				form.optional.data = '0'
			if item.timeout:
				form.timeout.data = f'{item.timeout}'
			else:
				form.timeout.data = ''
			form.release.data = item.release or ''
			form.value.data = item.value or ''
			form.value2.data = item.value2 or ''
			form.comparison.data = utils.getComparisonId(item.comparison)

	else:
		app.logger.warning(f'Item type "{item_type}" does not exist. Requested by "{current_user}".')
		flash(f'Item type "{item_type}" does not exist.', 'warning')
		return redirect(url_for('index'))

	if request.method == 'GET':
		form.name.data = item.name
		form.description.data = item.description

	if  form.validate_on_submit():
		# update item data
		item.name = form.name.data
		item.description = form.description.data
		# testrun
		if item_type == 'testrun':
			item.editor = current_user
			item.edited = datetime.utcnow()
			item.testcase_sequences=[models.TestCaseSequence.query.get(uuid.UUID(x).bytes) for x in form.testcase_sequences.data]
			item.global_settings = form.global_settings.data or None
		# testcase sequence
		elif item_type == 'testcase_sequence':
			item.edited = datetime.utcnow()
			item.classname = models.ClassName.query.get(int(form.classname.data))
			item.datafiles = [models.DataFile.query.get(uuid.UUID(x).bytes) for x in form.datafiles.data]
			item.testcases = [models.TestCase.query.get(uuid.UUID(x).bytes) for x in form.testcases.data]
		# testcase
		elif item_type == 'testcase':
			item.editor = current_user
			item.edited = datetime.utcnow()
			item.classname = models.ClassName.query.get(int(form.classname.data))
			item.browser_type = models.BrowserType.query.get(int(form.browser_type.data))
			item.testcase_type = models.TestCaseType.query.get(int(form.testcase_type.data))
			item.teststep_sequences = [models.TestStepSequence.query.get(uuid.UUID(x).bytes) for x in form.teststep_sequences.data]
		# step sequence
		elif item_type == 'teststep_sequence':
			item.editor = current_user
			item.edited = datetime.utcnow()
			item.classname = models.ClassName.query.get(int(form.classname.data))
			item.teststeps = [models.TestStepExecution.query.get(uuid.UUID(x).bytes) for x in form.teststeps.data]
		# test step
		elif item_type == 'teststep':
			item.editor = current_user
			item.edited = datetime.utcnow()
			item.activity_type = models.ActivityType.query.get(int(form.activity_type.data))
			item.locator_type = models.LocatorType.query.get(int(form.locator_type.data))
			# model extension
			item.locator = form.locator.data
			item.optional = [False, True][int(form.optional.data)]
			try:
				item.timeout = float(form.timeout.data)
			except ValueError:
				item.timeout = None
			item.release = form.release.data or None
			item.value = form.value.data or None
			item.value2 = form.value2.data or None
			if form.comparison.data == '0':
				item.comparison = None
			else:
				item.comparison = utils.COMPARISONS[int(form.comparison.data)]
			

		# update item in db
		db.session.commit()
		app.logger.info(f'Edited {item_type} {item_id} by {current_user}.')
		flash(f'Item "{item.name}" successfully updated.', 'success')
		return redirect(url_for('item_list', item_type=item_type))


	return render_template('testrun/edit_item.html', type=item_type, item=item, form=form)


@app.route('/<string:item_type>/new', methods=['GET', 'POST'])
@login_required
def new_item(item_type):
	#
	# create new item
	#
	if item_type == 'testrun':
		form = forms.TestrunCreateForm.new()
		chips = ['testcase_sequences']
	elif item_type == 'testcase_sequence':
		form = forms.TestCaseSequenceCreateForm.new(defaults=(request.method == 'GET'))
		chips = ['datafiles', 'testcases']
	elif item_type == 'testcase':
		form = forms.TestCaseCreateForm.new(defaults=(request.method == 'GET'))
		chips = ['teststep_sequences']
	elif item_type == 'teststep_sequence':
		form = forms.TestStepSequenceCreateForm.new(defaults=(request.method == 'GET'))
		chips = ['teststeps']
	elif item_type == 'teststep':
		form = forms.TestStepCreateForm.new(defaults=(request.method == 'GET'))
		chips = []
	else:
		app.logger.warning(f'Item type "{item_type}" does not exist. Requested by "{current_user}".')
		flash(f'Item type "{item_type}" does not exist.', 'warning')
		return redirect(url_for('index'))

	if form.validate_on_submit():
		# create new item
		# testrun
		if item_type == 'testrun':
			item = models.Testrun(
				name=form.name.data,
				description=form.description.data,
				creator=current_user,
				testcase_sequences=[models.TestCaseSequence.query.get(uuid.UUID(x).bytes) for x in form.testcase_sequences.data],
				global_settings=form.global_settings.data or None,
			)
		# testcase sequence
		elif item_type == 'testcase_sequence':
			item = models.TestCaseSequence(
				name=form.name.data,
				description=form.description.data,
				creator=current_user,
				classname=models.ClassName.query.get(int(form.classname.data)),
				datafiles=[models.DataFile.query.get(uuid.UUID(x).bytes) for x in form.datafiles.data],
				testcases=[models.TestCase.query.get(uuid.UUID(x).bytes) for x in form.testcases.data],
			)
		# testcase
		elif item_type == 'testcase':
			item = models.TestCase(
				name=form.name.data,
				description=form.description.data,
				creator=current_user,
				classname=models.ClassName.query.get(int(form.classname.data)),
				browser_type=models.BrowserType.query.get(int(form.browser_type.data)),
				testcase_type=models.TestCaseType.query.get(int(form.testcase_type.data)),
				teststep_sequences=[models.TestStepSequence.query.get(uuid.UUID(x).bytes) for x in form.teststep_sequences.data],
			)
		# step sequence
		elif item_type == 'teststep_sequence':
			item = models.TestStepSequence(
				name=form.name.data,
				description=form.description.data,
				creator=current_user,
				classname=models.ClassName.query.get(int(form.classname.data)),
				teststeps=[models.TestStepExecution.query.get(uuid.UUID(x).bytes) for x in form.teststeps.data],
			)
		# test step
		elif item_type == 'teststep':
			item = models.TestStepExecution(
				name=form.name.data,
				description=form.description.data,
				creator=current_user,
				activity_type=models.ActivityType.query.get(int(form.activity_type.data)),
				locator_type=models.LocatorType.query.get(int(form.locator_type.data)),
				# model extension
				locator=form.locator.data or None,
				optional=[False, True][int(form.optional.data)],
			)
			try:
				item.timeout = float(form.timeout.data)
			except ValueError:
				item.timeout = None
			item.release = form.release.data or None
			item.value = form.value.data or None
			item.value2 = form.value2.data or None
			if form.comparison.data == '0':
				item.comparison = None
			else:
				item.comparison = utils.COMPARISONS[int(form.comparison.data)]

		# save item to db
		db.session.add(item)
		db.session.commit()
		app.logger.info(f'Created {item_type} {item.uuid} by {current_user}.')
		flash(f'Item "{item.name}" successfully created.', 'success')
		return redirect(url_for('item_list', item_type=item_type))

	return render_template('testrun/create_item.html', type=item_type, chips=chips, form=form)


@app.route('/testrun/<string:item_id>/run')
@login_required
def run_testrun(item_id):
	#
	# runs Testrun via API Web Service
	#

	# check for complete set of Testrun's layers
	try:
		utils.getTestrunIfCorrect(item_id)
	except Exception as error:
		app.logger.error(f'Failed to execute Testrun {item_id} by {current_user}. {error}.')
		flash(f'ERROR: Testrun {item_id} cannot be executed. {error}', 'danger')
		return redirect(url_for('item_list', item_type='testrun'))
	 
	# export to JSON
	try:
		json_testrun = utils.exportJSON(item_id, shouldBeSaved=False)
	except Exception as error:
		app.logger.error(f'Failed to export Testrun {item_id} by {current_user}. {error}')
		flash(f'ERROR: Failed to export Testrun {item_id}. {error}', 'danger')
		return redirect(url_for('item_list', item_type='testrun'))

	# call API
	url = '/'.join(('http:/', app.config.get('BAANGT_API_HOST'), app.config.get('BAANGT_API_EXECUTE')))
	app.logger.info(f'Call execution of Testrun {item_id} by {current_user}. API URL: {url}')
	try:
		r = requests.post(url, json=json_testrun)
	except Exception as error:
		app.logger.error(f'Failed to connect to {app.config.get("BAANGT_API_HOST")} by {current_user}. {error}')
		flash(f'ERROR: Cannot establish connection: {app.config.get("BAANGT_API_HOST")}', 'danger')
		return redirect(url_for('item_list', item_type='testrun'))

	# API server error handler
	if r.status_code == 500:
		app.logger.error('ERROR: API Server internal error')
		flash('ERROR: API Server internal error', 'danger')
		return redirect(url_for('item_list', item_type='testrun'))

	app.logger.info(f'API response: status code {r.status_code}\n {r.text}')

	if r.status_code == 202:
		# get results
		jsonResult = json.loads(r.text)
		# store to db
		tr_call = models.TestRunCall(
			id=uuid.UUID(jsonResult['id']).bytes,
			creator=current_user,
			testrun_id=utils.idAsBytes(item_id),
		)
		db.session.add(tr_call)
		db.session.commit()
		app.logger.info(f'Created Testrun Call {jsonResult["id"]} by {current_user}.')
		return render_template('testrun/testrun_status.html', call=tr_call) 
	else:
		flash(f'ERROR: {r.text}', 'danger')
		return redirect(url_for('item_list', item_type='testrun'))


@app.route('/testrun/<string:item_id>/export/<string:export_format>')#, methods=['GET', 'POST'])
@login_required
def export_testrun(item_id, export_format):
	#
	# export Testrun object to <export_format>
	# <export_format> is one of the following:
	#   XLSX
	#   JSON
	#

	if export_format.upper() == 'XLSX':
		# export to XLSX
		try:
			result = utils.exportXLSX(item_id)
		except Exception as e:
			app.logger.error(f'Failed to export Testrun {item_id} to XLSX by {current_user}. {e}')
			return jsonify({'error': f'Failed to export Testrun: {e}'}), 400

	elif export_format.upper() == 'JSON':
		# export to JSON
		try:
			result = utils.exportJSON(item_id)
		except Exception as e:
			app.logger.error(f'Failed to export Testrun {item_id} to JSON by {current_user}. {e}')
			return jsonify({'error': f'Failed to export Testrun: {e}'}), 400

	else:
		# invalid format
		app.logger.error(f'Failed to export Testrun {item_id} by {current_user}. Invalid format: {export_format}')
		return jsonify({'error': f'format {export_format} is not supported'}), 400

	# successful export
	url = url_for('static', filename=f'files/{result}')
	app.logger.info(f'Testrun {item_id} exported to {export_format} by {current_user}. Target: static/files/{result}')

	#return f'Success: <a href="{url}">{result}</a>'
	return jsonify({
		'success': 'OK',
		'filename': result,
		'url': url,
	}), 200 

@app.route('/testrun/import', methods=['POST'])
@login_required
def import_testsun():
	#
	# imports testrun from file
	#

	# import only from XLSX is available now
	form = forms.TestrunImportForm()

	if form.validate_on_submit():
		try:
			importTestrun(form.file.data, datafiles=form.dataFiles)
			app.logger.info(f'Testrun successfully imported from "{form.file.data.filename}" by {current_user}.')
			flash(f'Testrun successfully imported from "{form.file.data.filename}"', 'success')
		except Exception as error:
			# discard changes
			db.session.rollback()
			app.logger.error(f'Failed to import Testrun from "{form.file.data.filename}" by {current_user}. {error}.')
			flash(f'ERROR: Cannot import Testrun from "{form.file.data.filename}". {error}.', 'danger')
	else:
		flash(f'File is required for import', 'warning')

	return redirect(url_for('item_list', item_type='testrun'))

@app.route('/datafile/upload', methods=['POST'])
@login_required
def upload_datafile():
	# get datafile
	file = request.files.get('datafile')
	if file is None:
		return jsonify({'error': 'Request does not contain dataFile'}), 400

	# upload data file
	try:
		file_id = utils.importDataFile(file)
	except Exception as error:
		app.logger.error(f'Failed to uplod DataFile "{file.filename}" by {current_user}. {error}.')
		return jsonify({'error': f'Failed upload DataFile "{file.filename}". {error}.'}), 400
	
	return jsonify({'id': file_id, 'name': file.filename}), 200

@app.route('/datafile/update', methods=['POST'])
@login_required
def update_datafile():
	#
	# update datafile with specified id
	#

	form = forms.DataFileUpdateForm()

	if form.validate_on_submit():
		file = form.datafile.data

		# upload data file
		try:
			file_id = utils.importDataFile(
				file,
				datafile_id=form.datafile_id.data,
			)
		except Exception as error:
			app.logger.error(f'Failed to uplod DataFile "{file.filename}" by {current_user}. {error}.')
			flash(f'ERROR: Failed to upload DataFile "{file.filename}". {error}.', 'danger')
			return redirect(url_for('item_list', item_type='testcase_sequence'))

		return redirect(url_for(
			'item_list',
			item_type='testcase_sequence',
			show=form.testcaseseq_id.data,
			_anchor=f'item{form.testcaseseq_id.data}',
		))

	app.logger.error(f'Failed to upload DataFile: {form.datafile.errors}')
	flash(f'ERROR: Failed to upload DataFile. {form.datafile.errors}.', 'danger')
	return redirect(url_for('item_list', item_type='testcase_sequence'))


@app.route('/testrun/<string:item_id>/import', methods=['POST'])
@login_required
def update_testsun(item_id):
	#
	# imports testrun from file
	#

	# update only from XLSX is available now
	form = forms.TestrunImportForm()

	if form.validate_on_submit():
		
		# update items
		try:
			importTestrun(form.file.data, testrun_id=item_id)
			app.logger.info(f'Testrun {item_id} successfully updated from "{form.file.data.filename}" by {current_user}.')
			flash(f'Testrun {item_id} has been successfully updated from "{form.file.data.filename}"', 'success')
		except Exception as error:
			# discard changes
			db.session.rollback()
			app.logger.error(f'Failed to update Testrun {item_id} from "{form.file.data.filename}" by {current_user}. {error}.')
			flash(f'ERROR: Cannot update Testrun {item_id} from "{form.file.data.filename}". {error}.', 'danger')
	else:
		flash(f'File is required for import', 'warning')

	return redirect(url_for('item_list', item_type='testrun'))


#
# Testrun Reports
#

@app.route('/summary/<string:call_id>', methods=['GET', 'POST'])
@login_required
def get_results(call_id):
	#
	# show Testrun call results
	#

	# fetch call data
	if request.method == 'POST':

		# call API
		url = '/'.join(('http:/', app.config.get('BAANGT_API_HOST'), app.config.get('BAANGT_API_STATUS'), call_id))
		app.logger.info(f'Call results by {current_user}. API URL: {url}')
		try:
			r = requests.get(url)
		except Exception as error:
			app.logger.error(f'Failed to connect to {app.config.get("TESTRUN_SERVICE_HOST")} by {current_user}. {error}')
			return jsonify({'error': f'Cannot establish connection: {app.config.get("TESTRUN_SERVICE_HOST")}'}), 400

		app.logger.info(f'API response: status code {r.status_code} -  JSON {r.text}')

		# API server error handler
		if r.status_code == 500:
			app.logger.error('ERROR: API Server internal error')
			return jsonify({'error': 'API Server internal error'}), 400

		if r.status_code == 200:
			# build charts
			try:
				charts = SummaryCharts(json.loads(r.text))
				return jsonify(charts.get_charts()), 200
			except Exception as e:
				app.logger.error(f'ERROR: {e}')
				return jsonify({'error': f'{e}'}), 400

		return jsonify({'error': f'Response: {r.status_code}'}), 400

	# show results
	call = models.TestRunCall.query.get(uuid.UUID(call_id).bytes)
	if not call:
		abort(404)

	if not call.is_finished and not call.is_failed:
		# update Testrun calls
		try:
			utils.update_call(call)
		except Exception as e:
			flash(f'{e}', 'warning')

	if not call.is_finished or call.is_failed:
		return render_template('testrun/testrun_status.html', call=call)

	return render_template('testrun/summary.html', call=call)


@app.route('/chart/<string:testrun_id>/<string:stage>', methods=['POST'])
@login_required
def get_charts(testrun_id, stage):
	#
	# fetches chart data
	#

	testrun = models.Testrun.query.get(uuid.UUID(testrun_id).bytes)

	# build charts
	try:
		charts = DashboardCharts(testrun.finished_calls(stage=stage))
		return jsonify(charts.get_charts()), 200
	except Exception as e:
		print(f'ERROR: {e}')
		return jsonify({'error': 'error'}), 400


@app.route('/dashboard')
@login_required
def dashboard():
	#
	# Dashboard view
	#

	# update Testrun calls
	try:
		utils.update_all_calls()
	except Exception as e:
		flash(f'{e}', 'warning')

	testruns = [tr for tr in models.Testrun.query.order_by(func.lower(models.Testrun.name)).all() if tr.finished_calls()]
	stages = []
	stage_set = set()
	for tr in testruns:
		stages.append({call.stage for call in tr.calls if call.is_finished})
		stage_set.update(stages[-1])
	
	return render_template('testrun/dashboard.html', testruns=testruns, stages=stages, stage_set=stage_set)


#
# user authentication
#

@app.route('/signup', methods=['GET', 'POST'])
def register():
	if current_user.is_authenticated:
		return redirect(url_for('index'))

	form = forms.SingupForm()
	if form.validate_on_submit():
		# create user
		user = models.User(username=form.username.data)
		user.set_password(form.password.data)
		db.session.add(user)
		db.session.commit()
		# login
		login_user(user, remember=True)
		flash(f'User {user.username.capitalize()} successfully created!', 'success')
		app.logger.info(f"User '{user.username}' successfully created")
		return redirect(url_for('index'))

	return render_template('auth/signup.html', form=form)

@app.route('/login', methods=['GET', 'POST'])
def login():
	if current_user.is_authenticated:
		return redirect(url_for('index'))

	form = forms.LoginForm()
	if form.validate_on_submit():
		user = models.User.query.filter_by(username=form.username.data).first()
		if user and user.verify_password(form.password.data):
			login_user(user, remember=True)
			flash(f'Welcome {user.username.capitalize()}!', 'success')
			app.logger.info(f"User '{user.username}' logged in")
			return redirect(url_for('index'))

	return render_template('auth/login.html', form=form)

@app.route('/logout')
@login_required
def logout():
	app.logger.info(f"User '{current_user.username}' logged out")
	logout_user()

	return redirect(url_for('login'))
	

#
# BAANGT documentations
#

@app.route('/docs/')
def help_index():
	return render_template('docs/index.html')

@app.route('/docs/<string:topic>')
def help(topic):
	if topic == 'testrun':
		try:
			with open(app.config.get('BAANGT_TESTRUN_GLOBALS_DEFAULT'), 'r') as f:
				global_settings = json.load(f)
			return render_template('docs/index.html', topic=topic, globals=json.dumps(global_settings, indent=2))
		except Exception as e:
			app.logger.error(f'Failed to load Testrun global settings: {e}')

	return render_template('docs/index.html', topic=topic)


@app.route('/testrun/<string:item_id>/json')
@login_required
def testrun_json(item_id):
	#
	# runs Testrun via API Web Service
	#

	# check for complete set of Testrun's layers
	try:
		utils.getTestrunIfCorrect(item_id)
	except Exception as error:
		app.logger.error(f'Failed to execute Testrun {item_id} by {current_user}. {error}.')
		flash(f'ERROR: Testrun {item_id} cannot be executed. {error}', 'danger')
		return redirect(url_for('item_list', item_type='testrun'))

	# export to JSON
	try:
		json_testrun = utils.exportJSON(item_id, shouldBeSaved=False)
	except Exception as error:
		app.logger.error(f'Failed to export Testrun {item_id} by {current_user}. {error}')
		flash(f'ERROR: Failed to export Testrun {item_id}. {error}', 'danger')
		return redirect(url_for('item_list', item_type='testrun'))
	return jsonify(json_testrun), 200

@app.route('/testrun/<string:item_id>/runlocal', methods=["POST"])
@login_required
def testrun_runlocal(item_id):
	jsonResult = request.get_json()
	# store to db
	tr_call = models.TestRunCall(
		id=uuid.UUID(jsonResult['id']).bytes,
		creator=current_user,
		testrun_id=utils.idAsBytes(item_id),
	)
	db.session.add(tr_call)
	db.session.commit()
	app.logger.info(f'Created Testrun Call {jsonResult["id"]} by {current_user}.')
	return jsonify({"success": 'true'}), 200

@app.route('/testrun/importdirect', methods=['POST'])
@login_required
def import_testrundirect():
	#
	# imports testrun from file
	#

	# import only from XLSX is available now
	try:
		data = request.files['file']
		importTestrun(data)
		app.logger.info(f'Testrun successfully imported from "{data.filename}" by {current_user}.')
		flash(f'Testrun successfully imported from "{data.filename}"', 'success')
	except Exception as error:
		# discard changes
		db.session.rollback()
		app.logger.error(f'Failed to import Testrun from "{data.filename}" by {current_user}. {error}.')
		flash(f'ERROR: Cannot import Testrun from "{data.filename}". {error}.', 'danger')
		return jsonify({"error": str(error)}), 200

	return redirect(url_for('item_list', item_type='testrun'))

