import pytest
from tests.supports import login, get_csrf, generate_password
from app import models
from lxml import html


#
# Register
#
def test_get_register(client, test_user):
    # url
    url = '/signup'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 200, f'Unauthorized access. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="signup-form"]')
    assert len(form) == 1, f'Sign Up Form absents'
    form = form[0]
    input_fields = form.xpath('./input')
    assert len(input_fields) == 4, f'Signup form has {len(input_fields)} input fields. Should be 4'

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # username
    field = form.xpath('.//input[@id="username"]')
    assert bool(field), 'Username field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, f'Username field is NOT empty'
    # password
    field = form.xpath('.//input[@id="password"]')
    assert bool(field), 'Password field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, f'Password field is NOT empty'
    assert field[0].attrib.get('type') == 'password', 'Password is NOT masked'
    # password confirmation
    field = form.xpath('.//input[@id="password2"]')
    assert bool(field), 'Password Confirmation field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, f'Password Confirmation field is NOT empty'
    assert field[0].attrib.get('type') == 'password', 'Password conformation is NOT masked'

    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]')), 'Submit button absents'

    # link to login
    assert bool(page.xpath('//a[@href="/login"]')), 'Link to Login Page absents'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 302, f'Authorized access. Status code: {r.status_code}'

def test_post_register(client, test_user):
    # url
    url = '/signup'

    user_data = {
        'username': 'SignUpTestUser',
        'password': generate_password(),
    }
    user_data['password2'] = user_data.get('password')

    # signup without CSRF
    r = client.post(url, data=user_data)
    assert r.status_code == 200, f'Sing up WITHOUT CSRF. Status code: {r.status_code}'
    # assert signup page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="signup-form"]')), f'Sign up WITHOUT CSRF. Signup Form absents'
    assert not bool(page.xpath('//form[@id="signup-form"]/p[contains(@class, "text-danger")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # signup existing account
    user_data={
        'csrf_token': get_csrf(client, url),
        'username': test_user.get('username'),
        'password': test_user.get('password'),
        'password2': test_user.get('password'),
    }
    r = client.post(url, data=user_data)
    assert r.status_code == 200, f'Sign up EXISTING user. Status code: {r.status_code}'
    # assert signup page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="signup-form"]')), f'Sign up EXISTING user. Signup form absents'
    assert bool(page.xpath('//form[@id="signup-form"]/p[contains(@class, "text-danger")]')), 'No error message for sign up EXISTING user'

    # signup without password
    user_data={
        'csrf_token': get_csrf(client, url),
        'username': 'SignUpTestUser',
    }
    r = client.post(url, data=user_data)
    assert r.status_code == 200, f'Sign up WITHOUT PASSWORD. Status code: {r.status_code}'
    # assert signup page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="signup-form"]')), f'Sign up WITHOUT PASSWORD. Signup form absents'
    assert len(page.xpath('//form[@id="signup-form"]/p[contains(@class, "text-danger")]')) == 2, \
        'No error message for sign up WITHOUT PASSWORD'

    # signup with wrong password confirmation
    user_data['password'] = generate_password()
    user_data['password2'] = generate_password()
    r = client.post(url, data=user_data)
    assert r.status_code == 200, f'Sign up with WRONG PASSWORD confirmation. Status code: {r.status_code}'
    # assert signup page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="signup-form"]')), f'Sign up with WRONG PASSWORD confirmation. Signup form absents'
    assert len(page.xpath('//form[@id="signup-form"]/p[contains(@class, "text-danger")]')) == 1, \
        'No error message for sign up with WRONG PASSWORD confirmation'

    # correct signup
    user_data['password2'] = user_data.get('password')
    r = client.post(url, data=user_data)
    assert r.status_code == 302, f'Sign up with CORRECT data. Status code: {r.status_code}'
    
    # assert created user
    user = models.User.query.filter_by(username=user_data.get('username')).first()
    assert bool(user), 'Created user was not saved to db'
    assert user.verify_password(user_data.get('password')), 'Created User password does not match'


#
# Login
#
def test_get_login(client, test_user):
    # url
    url = '/login'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 200, f'Unauthorized access. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="login-form"]')
    assert len(form) == 1, f'Login Form absents'
    form = form[0]
    input_fields = form.xpath('./input')
    assert len(input_fields) == 3, f'Login form has {len(input_fields)} input fields. Should be 3'

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # username
    field = form.xpath('.//input[@id="username"]')
    assert bool(field), 'Username field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, f'Username field is NOT empty'
    # password
    field = form.xpath('.//input[@id="password"]')
    assert bool(field), 'Password field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, f'Password field is NOT empty'
    assert field[0].attrib.get('type') == 'password', 'Password is NOT masked'

    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]')), 'Submit button absents'

    # link to signup
    assert bool(page.xpath('//a[@href="/signup"]')), 'Link to Login Page absents'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 302, f'Authorized access. Status code: {r.status_code}'

def test_post_login(client, test_user):
    # url
    url = '/login'

    user_data = {
        'username': test_user.get('username'),
        'password': test_user.get('password'),
    }

    # login without CSRF
    r = client.post(url, data=user_data)
    assert r.status_code == 200, f'Login WITHOUT CSRF. Status code: {r.status_code}'
    # assert login page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="login-form"]')), f'Login WITHOUT CSRF. Login Form absents'
    assert not bool(page.xpath('//form[@id="login-form"]/p[contains(@class, "text-danger")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

     # invalid user login
    user_data.update({
        'csrf_token': get_csrf(client, url),
        'username': 'TestLoginUser',
    })
    r = client.post(url, data=user_data)
    assert r.status_code == 200, f'INVALID user login. Status code: {r.status_code}'
    # assert login page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="login-form"]')), f'INVALID user login. Login form absents'
    assert bool(page.xpath('//form[@id="login-form"]/p[contains(@class, "text-danger")]')), 'No error message for INVALID user login'

    # login without password
    user_data={
        'csrf_token': user_data.get('csrf_token'),
        'username': test_user.get('username'),
    }
    r = client.post(url, data=user_data)
    assert r.status_code == 200, f'Login WITHOUT PASSWORD. Status code: {r.status_code}'
    # assert login page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="login-form"]')), f'Login WITHOUT PASSWORD. Login form absents'
    assert bool(page.xpath('//form[@id="login-form"]/p[contains(@class, "text-danger")]')), 'No error message for login WITHOUT PASSWORD'

    # login with wrong password
    user_data['password'] = generate_password()
    r = client.post(url, data=user_data)
    assert r.status_code == 200, f'Login with WRONG PASSWORD. Status code: {r.status_code}'
    # assert login page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="login-form"]')), f'Login with WRONG PASSWORD. Login form absents'
    assert bool(page.xpath('//form[@id="login-form"]/p[contains(@class, "text-danger")]')), 'No error message for login with WRONG PASSWORD'

    # correct login
    user_data['password'] = test_user.get('password')
    r = client.post(url, data=user_data)
    assert r.status_code == 302, f'Login with CORRECT data. Status code: {r.status_code}'
    

#
# Logout
#
def test_logout(client, test_user):
    # url
    url = '/logout'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 302, f'Authorized access. Status code: {r.status_code}'
