import pytest
from app import models
from tests.supports import login, path_to_data, get_csrf
from lxml import html
from random import choice
import json
import os
import re


@pytest.mark.parametrize('count', [(1,1)])
def test_get_results_success(client, test_user, populate, set_tr_calls):
    # set-up  
    with open(os.path.join(path_to_data, 'testrun_calls.json'), 'r') as f:
        calls = json.load(f)
    tr  = choice(models.Testrun.query.all())  
    set_tr_calls(tr.id, calls.get('success'))
    tr_call = choice(models.TestRunCall.query.all())
    urls = {
        'correct': f'/summary/{tr_call}',
        'fake': f'/summary/{tr.uuid}',
    }

    # unauthorized access
    r = client.get(urls.get('correct'))
    assert r.status_code == 302
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)

    # GET fake call
    r = client.get(urls.get('fake'))
    assert r.status_code == 404, f'GET fake call. Status code: {r.status_code}'

    # GET corret call    
    r = client.get(urls.get('correct'))
    assert r.status_code == 200, f'GET request. Status code: {r.status_code}'
    # assert summary page
    page = html.fromstring(r.data)
    main = page.xpath(f'//main[@id="{tr_call}"]')
    assert bool(main), 'Main section with TestRunCall ID is absent'

@pytest.mark.parametrize('count', [(1,1)])
def test_get_results_error(client, test_user, populate, set_tr_calls):
    # set-up  
    with open(os.path.join(path_to_data, 'testrun_calls.json'), 'r') as f:
        calls = json.load(f)
    tr  = choice(models.Testrun.query.all())  
    set_tr_calls(tr.id, calls.get('error'))
    tr_call = choice(models.TestRunCall.query.all())
    urls = {
        'correct': f'/summary/{tr_call}',
        'fake': f'/summary/{tr.uuid}',
    }

    # unauthorized access
    r = client.get(urls.get('correct'))
    assert r.status_code == 302
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)

    # GET fake call
    r = client.get(urls.get('fake'))
    assert r.status_code == 404, f'GET fake call. Status code: {r.status_code}'

    # GET corret call    
    r = client.get(urls.get('correct'))
    assert r.status_code == 200, f'GET request. Status code: {r.status_code}'
    # assert error page
    page = html.fromstring(r.data)
    execution_status = page.xpath('//div[@id="execution-error"]')
    assert bool(execution_status), 'Card with Testrun Execution Status is absent'
    execution_id = execution_status[0].xpath('./h5[@id="call-id"]')
    assert bool(execution_id)
    call_id_match = re.search(r'[\w-]+$', execution_id[0].text)
    assert bool(call_id_match), 'There is no call ID in error screen'
    assert call_id_match.group() == str(tr_call), 'Call ID does not match'


@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_get_charts(client, test_user, populate, set_tr_calls):
    # set-up  
    with open(os.path.join(path_to_data, 'testrun_calls.json'), 'r') as f:
        calls = json.load(f)
    tr  = choice(models.Testrun.query.all())
    set_tr_calls(tr.id, calls.get('success')) 
    set_tr_calls(tr.id, calls.get('error'))
    tr_call = choice(models.TestRunCall.query.all())
    stage = calls.get('success')[0]['stage']
    urls = {
        'correct': f'/chart/{tr.uuid}/{stage}',
        'fake_tr': f'/chart/{tr_call}/{stage}',
        'fake_stage': f'/chart/{tr.uuid}/fake'
    }

    # unauthorized access
    r = client.post(urls.get('correct'))
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)

    # GET request
    r = client.get(urls.get('correct'))
    assert r.status_code == 405, f'GET request. Status code: {r.status_code}'

    # POST fake testrun
    r = client.post(urls.get('fake_tr'))
    print(r.data)
    assert r.status_code == 400, f'POST fake testrun. Status code: {r.status_code}'

    # POST fake stage
    r = client.post(urls.get('fake_stage'))
    print(r.data)
    assert r.status_code == 400, f'GET fake call. Status code: {r.status_code}'

    # POST correct request
    r = client.post(urls.get('correct'))
    assert r.status_code == 200, f'GET request. Status code: {r.status_code}'
    # assert response
    response = json.loads(r.data)
    assert len(response) == 4, 'Response should contain 4 charts'
    
    # figures data should show the recent call
    figure = 'figures'
    fields = [
        'records',
        'successful',
        'error',
        'paused',
    ]
    assert figure in response, f"Data for '{figure}' is absent"
    assert all(f in response.get(figure) for f in fields), f"Not all fields in '{figure}'"
    # get the recent call
    recent_call = None
    for call in calls.get('success'):
        if call.get('stage') == stage:
            recent_call = call
    for field in fields:
        attr = field
        if field == 'records':
            attr = 'testrecords'
        assert response.get(figure)[field] == recent_call.get(attr), f"Field '{field}' in '{figure}' does not match"

    # status data should show the recent call
    figure = 'status'
    labels = {
        'Passed': 'successful',
        'Failed': 'error',
        'Paused': 'paused',
    }
    assert figure in response, f"Data for '{figure}' is absent"
    assert response.get(figure).get('type') == 'doughnut', f"Chart '{figure}' is NOT DOUGHNUT"
    assert 'data' in response.get(figure), f"Data is absent for '{figure}'"
    assert 'datasets' in response.get(figure)['data'], f"Datasets are absent for '{figure}'"
    assert bool(response.get(figure)['data']['datasets']), f"Datasets are absent for '{figure}'"
    assert 'data' in response.get(figure)['data']['datasets'][0], f"Dataset is absent for '{figure}'"
    assert 'labels' in response.get(figure)['data'], f"labels are absent for '{figure}'"
    # get data
    chart_data = {d[0]: d[1] for d in zip(response.get(figure)['data']['labels'], response.get(figure)['data']['datasets'][0]['data'])}
    assert chart_data == {l: recent_call.get(labels.get(l)) for l in labels}

    # duration
    figure = 'duration'
    assert figure in response, f"Data for '{figure}' is absent"
    assert response.get(figure).get('type') == 'line', f"Chart '{figure}' is NOT LINE"
    assert 'data' in response.get(figure), f"Data is absent for '{figure}'"
    assert 'datasets' in response.get(figure)['data'], f"Datasets are absent for '{figure}'"
    assert bool(response.get(figure)['data']['datasets']), f"Datasets are absent for '{figure}'"
    assert 'data' in response.get(figure)['data']['datasets'][0], f"Dataset is absent for '{figure}'"
    # get durations
    call_durations = zip(
        [call.get('duration') for call in calls.get('success') if call.get('stage') == stage],
        response.get(figure)['data']['datasets'][0]['data']
    )
    assert all(d[0] == d[1] for d in call_durations)

    # results
    figure = 'results'
    labels = {
        'Passed': 'successful',
        'Failed': 'error',
        'Paused': 'paused',
    }
    assert figure in response, f"Data for '{figure}' is absent"
    assert response.get(figure).get('type') == 'bar', f"Chart '{figure}' is NOT BAR"
    assert 'data' in response.get(figure), f"Data is absent for '{figure}'"
    assert 'datasets' in response.get(figure)['data'], f"Datasets are absent for '{figure}'"
    assert bool(response.get(figure)['data']['datasets']), f"Datasets are absent for '{figure}'"
    # get data
    for index, dataset in enumerate(response.get(figure)['data']['datasets']):
        assert 'data' in dataset, f"Dataset is absent for dataset {index} in '{figure}'"
        assert 'label' in dataset, f"Label is absent for dataset {index} in '{figure}'"
        call_data = zip(
            [call.get(labels.get(dataset.get('label'))) for call in calls.get('success') if call.get('stage') == stage],
            dataset['data']
        )
        assert all(d[0] == d[1] for d in call_data)


@pytest.mark.parametrize('count', [(0,0), (1,2), (5,2)])
def test_dashboard(client, test_user, populate, set_tr_calls):
    # set-up
    url = '/dashboard'
    with open(os.path.join(path_to_data, 'testrun_calls.json'), 'r') as f:
        calls = json.load(f)
    stages = {'All'}
    if models.Testrun.query.count() > 0:
        stages = {call.get('stage') for call in calls.get('success')}
        stages.add('All')
    # add calls to every testrun
    testruns = [('0', 'All')]
    for tr in models.Testrun.query.all():
        set_tr_calls(tr.id, calls.get('success')) 
        set_tr_calls(tr.id, calls.get('error'))
        testruns.append((tr.uuid, tr.name))
    
    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'GET request. Status code: {r.status_code}'
    # assert dashbord page elements
    page = html.fromstring(r.data)
    
    # filter by name
    name_field = page.xpath(f'//select[@id="selectName"]')
    assert bool(name_field), "Field 'Filter by Name' is absent"
    assert [(opt.attrib.get('value'), opt.text) for opt in name_field[0].xpath('./option')] == testruns, \
        'Testrun list in name-filter does NOT match'
    
    # filter by stage
    stage_field = page.xpath(f'//select[@id="selectStage"]')
    assert bool(stage_field), "Field 'Filter by Name' is absent"
    assert set(opt.text for opt in stage_field[0].xpath('./option')) == stages, 'Stage list in stage-filter does NOT match'

    # call elements
    call_elements = page.xpath(f'//div[contains(@class, "hovered")]')
    assert len(call_elements) == (len(testruns)-1)*(len(stages)-1), 'Number of calls does NOT match'
    
    
    



##### TODO:
# POST: get_results

