import pytest
from tests.supports import login, get_csrf
from app import models
from lxml import html
from random import choice, sample, randrange
import uuid


#
# Testrun Forms
#
@pytest.mark.parametrize('count', [(3,1), (2,3)])
def test_get_create_testrun(client, test_user, populate):
    # set support variables
    url = f'/testrun/new'
    sub_name = 'testcase_sequences'
    avail_sub_items_set = set(i.uuid for i in models.TestCaseSequence.query.all())

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'Get Create Testrun. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="create-form"]')
    assert len(form) == 1, f'Create Form absents'
    form = form[0]

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, f'Name field is NOT empty'
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, f"Description field is NOT empty"
    # global settings
    field = form.xpath('.//textarea[@id="global_settings"]')
    assert bool(field), 'GlobalSettings field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, 'Global settings value is NOT empty'
    # hidden multi-select field
    field = form.xpath(f'.//select[@id="{sub_name}"]')
    assert bool(field), 'Hidden TestCaseSequence select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_sub_items_set, \
        'Hidden TestCaseSequence select field holds NOT ALL items'
    # visible select field
    assert bool(form.xpath(f'.//input[@data-name="{sub_name}"]')), 'TestCaseSequences field absents'
    # chip area
    chip_area = form.xpath(f'.//div[@id="chips_{sub_name}"]')
    assert bool(chip_area), 'TestCaseSequence chip area absents'
    assert not bool(chip_area[0].text.strip()), 'Chip area is not empty'
    # datalist
    datalist = form.xpath(f'.//datalist[@id="{sub_name}Opt"]')
    assert bool(datalist), 'TestCaseSequences data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_sub_items_set), 'TestCaseSequence datalist holds NOT ALL items'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_sub_items_set, 'Datalist holds NOT ALL items'

    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

@pytest.mark.parametrize('count', [(1,1), (5,2), (3,0)])
def test_post_create_testrun(client, test_user, populate):
    # login
    url = f'/testrun/new'
    login(client, test_user)

    # post empty form
    r = client.post(url, data={})
    assert r.status_code == 200, f'Post empty Create Form. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Create Form absents'
    assert len(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')) == 2, 'No error message for EMPTY form'

    item_data = {
        'name': 'New Test Testrun',
        'description': 'New Test Testrun Description',
    }

    # post create without CSRF
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create From WITHOUT CSRF. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Create Form absents'
    assert not bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake sub items
    item_data['csrf_token'] = get_csrf(client, url)
    item_data['testcase_sequences'] = [str(uuid.uuid4())]
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE sub items. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE sub items'

    # post not JSON global settings
    item_data['testcase_sequences'] = [x.uuid for x in sample(models.TestCaseSequence.query.all(), k=populate[1])]
    item_data['global_settings'] = 'New Global Settings'
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with WRONG Global Settings. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for WRONG Global Settings'

    # create item
    item_data['global_settings'] = '{"New": "Global Settings"}'
    r = client.post(url, data=item_data)
    assert r.status_code == 302, f'Post Create Form with CORRECT form. Status code: {r.status_code}'

    # assert created element
    item = models.Testrun.query.filter_by(name=item_data.get('name')).first()
    assert bool(item), 'Item NOT saved to DB'
    assert item.description == item_data.get('description')
    assert set(x.uuid for x in item.testcase_sequences) == set(item_data.get('testcase_sequences'))
    assert item.global_settings == item_data.get('global_settings')

@pytest.mark.parametrize('count', [(3,1), (2,3)])
def test_get_edit_testrun(client, test_user, populate):
    # get random testrun item
    item = choice(models.Testrun.query.all())
    url = f'/testrun/{item.uuid}/edit'
    sub_items_set = set(i.uuid for i in item.testcase_sequences)
    sub_name = 'testcase_sequences'
    avail_sub_items_set = set(i.uuid for i in models.TestCaseSequence.query.all())


    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'Get Edit Testrun. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="edit-form"]')
    assert len(form) == 1, f'Edit Form absents'
    form = form[0]

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = field[0].attrib.get('value')
    assert value == item.name, f"Name field value '{value}' does NOT match item name '{item.name}'"
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = field[0].attrib.get('value')
    assert value == item.description, f"Description field value '{value}' does NOT match item description '{item.description}'"
    # global settings
    field = form.xpath('.//textarea[@id="global_settings"]')
    assert bool(field), 'GlobalSettings field absents'
    value = field[0].attrib.get('value')
    assert value == str(item.global_settings), 'Global settings value does NOT match item global_settings'
    # hidden multi-select field
    field = form.xpath(f'.//select[@id="{sub_name}"]')
    assert bool(field), 'Hidden TestCaseSequence select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_sub_items_set, \
        'Hidden TestCaseSequence select field holds NOT ALL items'
    # visible select field
    assert bool(form.xpath(f'.//input[@data-name="{sub_name}"]')), 'TestCaseSequences field absents'
    # chip area
    chip_area = form.xpath(f'.//div[@id="chips_{sub_name}"]')
    assert bool(chip_area), 'TestCaseSequence chip area absents'
    chips = chip_area[0].xpath('./div')
    assert set(chip.attrib.get('data-id') for chip in chips) == sub_items_set
    # datalist
    datalist = form.xpath(f'.//datalist[@id="{sub_name}Opt"]')
    assert bool(datalist), 'TestCaseSequences data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_sub_items_set), 'TestCaseSequence datalist holds NOT ALL items'
    assert set(opt.attrib.get('data-id') for opt in options if opt.attrib.get('disabled')) == sub_items_set

    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

@pytest.mark.parametrize('count', [(1,1), (5,2), (3,0)])
def test_post_edit_testrun(client, test_user, populate):
    # get random testrun item
    item = choice(models.Testrun.query.all())
    url = f'/testrun/{item.uuid}/edit'

    # login
    login(client, test_user)
    update_data = {
        'name': 'Updated Testrun',
        'description': 'Updated Testrun Description',
    }

    # post update without CSRF
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit Testrun WITHOUT CSRF. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Edit Form absents'
    assert not bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake sub items
    update_data['csrf_token'] = get_csrf(client, url)
    update_data['testcase_sequences'] = [str(uuid.uuid4())]
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit Testrun with FAKE sub items. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE sub items'

    # post not JSON global settings
    update_data['testcase_sequences'] = [x.uuid for x in sample(models.TestCaseSequence.query.all(), k=populate[1])]
    update_data['global_settings'] = 'Update Global Settings'
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit Testrun with WRONG Global Settings. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for WRONG Global Settings'

    # update item
    update_data['global_settings'] = '{"Update": "Global Settings"}'
    r = client.post(url, data=update_data)
    assert r.status_code == 302, f'Post Edit Testrun with CORRECT form. Status code: {r.status_code}'

    # assert updated element
    item = models.Testrun.query.get(item.id)
    assert item.name == update_data.get('name')
    assert item.description == update_data.get('description')
    assert set(x.uuid for x in item.testcase_sequences) == set(update_data.get('testcase_sequences'))
    assert item.global_settings == update_data.get('global_settings')


#
# TestCaseSequence Forms
#
@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_get_create_testcase_sequence(client, test_user, populate, default_supports):
    # get random testcase sequence item
    url = f'/testcase_sequence/new'
    sub_name = 'testcases'
    avail_sub_items_set = set(i.uuid for i in models.TestCase.query.all())
    df_name = 'datafiles'
    avail_df_set = set(i.uuid for i in models.DataFile.query.all())

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'GET Create TestCaseSequence. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="create-form"]')
    assert len(form) == 1, f'Create Form absents'
    form = form[0]

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, f"Name field value is NOT empty"
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, f"Description field value is NOT empty"
    # classname
    field = form.xpath('.//select[@id="classname"]')
    assert bool(field), 'ClassName field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'ClassName field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.ClassName.query.all()), 'ClassName field holds NOT ALL items'
    assert [opt.text for opt in options if opt.attrib.get('selected')] == \
        [default_supports.get('CLASSNAME_TESTCASESEQUENCE')], 'Classname field default value is wrong'
    # hidden multi-select field
    field = form.xpath(f'.//select[@id="{sub_name}"]')
    assert bool(field), 'Hidden TestCase select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_sub_items_set, \
        'Hidden TestCase select field holds NOT ALL items'
    # visible select field
    assert bool(form.xpath(f'.//input[@data-name="{sub_name}"]')), 'TestCase field absents'
    # chip area
    chip_area = form.xpath(f'.//div[@id="chips_{sub_name}"]')
    assert bool(chip_area), 'TestCase chip area absents'
    assert not bool(chip_area[0].text.strip()), 'TestCase chip area is not empty'
    # datalist
    datalist = form.xpath(f'.//datalist[@id="{sub_name}Opt"]')
    assert bool(datalist), 'TestCase data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_sub_items_set), 'TestCase datalist holds NOT ALL items'
    assert set(opt.attrib.get('data-id') for opt in options) == avail_sub_items_set, 'Datalist holds NOT ALL items'
    # hidden datafile multi-select field
    field = form.xpath(f'.//select[@id="{df_name}"]')
    assert bool(field), 'Hidden DataFile select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_df_set, \
        'Hidden DataFile select field holds NOT ALL items'
    # visible datafile select field
    assert bool(form.xpath(f'.//input[@data-name="{df_name}"]')), 'DataFile field absents'
    # datafile chip area
    chip_area = form.xpath(f'.//div[@id="chips_{df_name}"]')
    assert bool(chip_area), 'DataFile chip area absents'
    assert not bool(chip_area[0].text.strip()), 'DataFile chip area is not empty'
    # datafile datalist
    datalist = form.xpath(f'.//datalist[@id="{df_name}Opt"]')
    assert bool(datalist), 'DataFile data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_df_set), 'DataFile datalist holds NOT ALL items'
    assert set(opt.attrib.get('data-id') for opt in options) == avail_df_set
    # datafile upload button
    assert bool(form.xpath('.//button[@id="datafile-upload"]')), 'DataFile Upload button absents'

    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_post_create_testcase_sequence(client, test_user, populate):
    # login
    url = f'/testcase_sequence/new'
    login(client, test_user)

    # post empty form
    r = client.post(url, data={})
    assert r.status_code == 200, f'Post empty Create Form. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Create Form absents'
    assert len(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')) == 3, 'No error message for EMPTY form'

    item_data = {
        'name': 'New Test TestCaseSequence',
        'description': 'New Test TestCaseSequence Description',
        'classname': 1,
    }

    # post update without CSRF
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form WITHOUT CSRF. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form WITHOUT CSRF. Create Form absents'
    assert not bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake classname items
    item_data['csrf_token'] = get_csrf(client, url)
    item_data['classname'] = 0
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Edit TestCaseSequence with FAKE classname. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form with FAKE classname. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE classname'

    # post fake sub items
    item_data['classname'] = choice(models.ClassName.query.all()).id
    item_data['testcases'] = [str(uuid.uuid4())]
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE sub items. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post reate Form with FAKE sub items. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE sub items'

    # post fake datafile
    item_data['testcases'] = [x.uuid for x in sample(models.TestCase.query.all(), k=populate[1])]
    item_data['datafiles'] = [str(uuid.uuid4())]
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE datafile. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form with FAKE datafile. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE datafile'

    # update item
    item_data['datafiles'] = [choice(models.DataFile.query.all()).uuid]
    r = client.post(url, data=item_data)
    assert r.status_code == 302, f'Post Create Form with CORRECT form. Status code: {r.status_code}'

    # assert created element
    item = models.TestCaseSequence.query.filter_by(name=item_data.get('name')).first()
    assert bool(item), 'Item NOT saved to DB'
    assert item.description == item_data.get('description')
    assert item.classname.id == item_data.get('classname')
    assert set(x.uuid for x in item.testcases) == set(item_data.get('testcases'))
    assert set(x.uuid for x in item.datafiles) == set(item_data.get('datafiles'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_get_edit_testcase_sequence(client, test_user, populate):
    # get random testcase sequence item
    item = choice(models.TestCaseSequence.query.all())
    url = f'/testcase_sequence/{item.uuid}/edit'
    sub_items_set = set(i.uuid for i in item.testcases)
    sub_name = 'testcases'
    avail_sub_items_set = set(i.uuid for i in models.TestCase.query.all())
    df_set = set(i.uuid for i in item.datafiles)
    df_name = 'datafiles'
    avail_df_set = set(i.uuid for i in models.DataFile.query.all())

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'GET Edit TestCaseSequence. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="edit-form"]')
    assert len(form) == 1, f'Edit Form absents'
    form = form[0]

    # reload item
    item = models.TestCaseSequence.query.get(item.id)

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = field[0].attrib.get('value')
    assert value == item.name, f"Name field value '{value}' does NOT match item name '{item.name}'"
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = field[0].attrib.get('value')
    assert value == item.description, f"Description field value '{value}' does NOT match item description '{item.description}'"
    # classname
    field = form.xpath('.//select[@id="classname"]')
    assert bool(field), 'ClassName field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'ClassName field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.ClassName.query.all()), 'ClassName field holds NOT ALL items'
    assert [opt.attrib.get('value') for opt in options if opt.attrib.get('selected')] == \
        [str(item.classname.id)], 'Classname field selected value is wrong'
    # hidden multi-select field
    field = form.xpath(f'.//select[@id="{sub_name}"]')
    assert bool(field), 'Hidden TestCase select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_sub_items_set, \
        'Hidden TestCase select field holds NOT ALL items'
    # visible select field
    assert bool(form.xpath(f'.//input[@data-name="{sub_name}"]')), 'TestCase field absents'
    # chip area
    chip_area = form.xpath(f'.//div[@id="chips_{sub_name}"]')
    assert bool(chip_area), 'TestCase chip area absents'
    chips = chip_area[0].xpath('./div')
    assert set(chip.attrib.get('data-id') for chip in chips) == sub_items_set
    # datalist
    datalist = form.xpath(f'.//datalist[@id="{sub_name}Opt"]')
    assert bool(datalist), 'TestCase data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_sub_items_set), 'TestCase datalist holds NOT ALL items'
    assert set(opt.attrib.get('data-id') for opt in options if opt.attrib.get('disabled')) == sub_items_set
    # hidden datafile multi-select field
    field = form.xpath(f'.//select[@id="{df_name}"]')
    assert bool(field), 'Hidden DataFile select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_df_set, \
        'Hidden DataFile select field holds NOT ALL items'
    # visible datafile select field
    assert bool(form.xpath(f'.//input[@data-name="{df_name}"]')), 'DataFile field absents'
    # datafile chip area
    chip_area = form.xpath(f'.//div[@id="chips_{df_name}"]')
    assert bool(chip_area), 'DataFile chip area absents'
    chips = chip_area[0].xpath('./div')
    assert set(chip.attrib.get('data-id') for chip in chips) == df_set
    # datafile datalist
    datalist = form.xpath(f'.//datalist[@id="{df_name}Opt"]')
    assert bool(datalist), 'DataFile data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_df_set), 'DataFile datalist holds NOT ALL items'
    assert set(opt.attrib.get('data-id') for opt in options if opt.attrib.get('disabled')) == df_set
    # datafile upload button
    assert bool(form.xpath('.//button[@id="datafile-upload"]')), 'DataFile Upload button absents'

    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_post_edit_testcase_sequence(client, test_user, populate):
    # get random testcase sequence item
    item = choice(models.TestCaseSequence.query.all())
    url = f'/testcase_sequence/{item.uuid}/edit'

    # login
    login(client, test_user)
    update_data = {
        'name': 'Updated TestCaseSequence',
        'description': 'Updated TestCaseSequence Description',
        'classname': 1,
    }

    # post update without CSRF
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestCaseSequence WITHOUT CSRF. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestCaseSequence WITHOUT CSRF. Edit Form absents'
    assert not bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake classname items
    update_data['csrf_token'] = get_csrf(client, url)
    update_data['classname'] = 0
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestCaseSequence with FAKE classname. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestCaseSequence with FAKE classname. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE classname'

    # post fake sub items
    update_data['classname'] = 1
    update_data['testcases'] = [str(uuid.uuid4())]
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestCaseSequence with FAKE sub items. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestCaseSequence with FAKE sub items. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE sub items'

    # post fake datafile
    update_data['testcases'] = [x.uuid for x in sample(models.TestCase.query.all(), k=populate[1])]
    update_data['datafiles'] = [str(uuid.uuid4())]
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestCaseSequence with FAKE datafile. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestCaseSequence with FAKE datafile. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE datafile'

    # update item
    update_data['datafiles'] = [choice(models.DataFile.query.all()).uuid]
    r = client.post(url, data=update_data)
    assert r.status_code == 302, f'Post Edit TestCaseSequence with CORRECT form. Status code: {r.status_code}'

    # assert updated element
    item = models.TestCaseSequence.query.get(item.id)
    assert item.name == update_data.get('name')
    assert item.description == update_data.get('description')
    assert item.classname.id == update_data.get('classname')
    assert set(x.uuid for x in item.testcases) == set(update_data.get('testcases'))
    assert set(x.uuid for x in item.datafiles) == set(update_data.get('datafiles'))


#
# TestCase Forms
#
@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_get_create_testcase(client, test_user, populate, default_supports):
    # set support vars
    url = f'/testcase/new'
    sub_name = 'teststep_sequences'
    avail_sub_items_set = set(i.uuid for i in models.TestStepSequence.query.all())

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'GET Create TestCase. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)
    form = page.xpath('//form[@id="create-form"]')
    assert len(form) == 1, f'Create Form absents'
    form = form[0]

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Name field value is NOT empty"
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Description field value is Not empty"
    # classname
    field = form.xpath('.//select[@id="classname"]')
    assert bool(field), 'ClassName field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'ClassName field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.ClassName.query.all()), 'ClassName field holds NOT ALL items'
    assert [opt.text for opt in options if opt.attrib.get('selected')] == \
        [default_supports.get('CLASSNAME_TESTCASE')], 'Classname field default value is wrong'
    # browser
    field = form.xpath('.//select[@id="browser_type"]')
    assert bool(field), 'Browser Type field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'Browser Type field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.BrowserType.query.all()), 'Browser Type field holds NOT ALL items'
    assert [opt.text for opt in options if opt.attrib.get('selected')] == \
        [default_supports.get('BROWSER_TYPE')], 'Browser Type field default value is wrong'
    # testcase type
    field = form.xpath('.//select[@id="testcase_type"]')
    assert bool(field), 'Testcase Type field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'Testcase Type field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.TestCaseType.query.all()), 'Testcase Type field holds NOT ALL items'
    assert [opt.text for opt in options if opt.attrib.get('selected')] == \
        [default_supports.get('TESTCASE_TYPE')], 'TestCaseType field default value is wrong'
    # hidden multi-select field
    field = form.xpath(f'.//select[@id="{sub_name}"]')
    assert bool(field), 'Hidden TestStepSequence select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_sub_items_set, \
        'Hidden TestStepSequence select field holds NOT ALL items'
    # visible select field
    assert bool(form.xpath(f'.//input[@data-name="{sub_name}"]')), 'TestStepSequence field absents'
    # chip area
    chip_area = form.xpath(f'.//div[@id="chips_{sub_name}"]')
    assert bool(chip_area), 'TestStepSequence chip area absents'
    assert not bool(chip_area[0].text.strip()), 'Chip area is not empty'
    # datalist
    datalist = form.xpath(f'.//datalist[@id="{sub_name}Opt"]')
    assert bool(datalist), 'TestStepSequence data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_sub_items_set), 'TestStepSequence datalist holds NOT ALL items'
    assert set(opt.attrib.get('data-id') for opt in options) == avail_sub_items_set, 'Datalist holds NOT ALL items'
    
    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_post_create_testcase(client, test_user, populate):
    # login
    url = f'/testcase/new'
    login(client, test_user)

    # post empty form
    r = client.post(url, data={})
    assert r.status_code == 200, f'Post empty Create Form. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Create Form absents'
    assert len(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')) == 5, 'No error message for EMPTY form'

    item_data = {
        'name': 'New Test TestCase',
        'description': 'New Test TestCase Description',
        'classname': 1,
        'browser_type': 1,
        'testcase_type': 1,
    }

    # post update without CSRF
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form WITHOUT CSRF. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form WITHOUT CSRF. Create Form absents'
    assert not bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake classname items
    item_data['csrf_token'] = get_csrf(client, url)
    item_data['classname'] = 0
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE classname. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Edit TestCase with FAKE classname. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE classname'

    # post fake browser type items
    item_data['classname'] = choice(models.ClassName.query.all()).id
    item_data['browser_type'] = 0
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE browser type. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form with FAKE browser type. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE browser type'

    # post fake testcase type items
    item_data['browser_type'] = choice(models.BrowserType.query.all()).id
    item_data['testcase_type'] = 0
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE testcase type. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form with FAKE testcase type. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE testcase type'

    # post fake subitems
    item_data['testcase_type'] = choice(models.TestCaseType.query.all()).id
    item_data['teststep_sequences'] = [str(uuid.uuid4())]
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE sub items. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form with FAKE sub items. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE sub items'

    # create item 
    item_data['teststep_sequences'] = [x.uuid for x in sample(models.TestStepSequence.query.all(), k=populate[1])]
    r = client.post(url, data=item_data)
    assert r.status_code == 302, f'Post Create Form with CORRECT form. Status code: {r.status_code}'

    # assert created element
    item = models.TestCase.query.filter_by(name=item_data.get('name')).first()
    assert bool(item), 'Item NOT saved to DB'
    assert item.description == item_data.get('description')
    assert item.classname.id == item_data.get('classname')
    assert item.browser_type.id == item_data.get('browser_type')
    assert item.testcase_type.id == item_data.get('testcase_type')
    assert set(x.uuid for x in item.teststep_sequences) == set(item_data.get('teststep_sequences'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_get_edit_testcase(client, test_user, populate):
    # get random testcase item
    item = choice(models.TestCase.query.all())
    url = f'/testcase/{item.uuid}/edit'
    sub_items_set = set(i.uuid for i in item.teststep_sequences)
    sub_name = 'teststep_sequences'
    avail_sub_items_set = set(i.uuid for i in models.TestStepSequence.query.all())

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'GET Edit TestCase. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="edit-form"]')
    assert len(form) == 1, f'Edit Form absents'
    form = form[0]

    # reload item
    item = models.TestCase.query.get(item.id)

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = field[0].attrib.get('value')
    assert value == item.name, f"Name field value '{value}' does NOT match item name '{item.name}'"
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = field[0].attrib.get('value')
    assert value == item.description, f"Description field value '{value}' does NOT match item description '{item.description}'"
    # classname
    field = form.xpath('.//select[@id="classname"]')
    assert bool(field), 'ClassName field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'ClassName field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.ClassName.query.all()), 'ClassName field holds NOT ALL items'
    assert [opt.attrib.get('value') for opt in options if opt.attrib.get('selected')] == \
        [str(item.classname.id)], 'Classname field selected value is wrong'
    # browser
    field = form.xpath('.//select[@id="browser_type"]')
    assert bool(field), 'Browser Type field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'Browser Type field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.BrowserType.query.all()), 'Browser Type field holds NOT ALL items'
    assert [opt.attrib.get('value') for opt in options if opt.attrib.get('selected')] == \
        [str(item.browser_type.id)], 'Browser Type field selected value is wrong'
    # testcase type
    field = form.xpath('.//select[@id="testcase_type"]')
    assert bool(field), 'Testcase Type field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'Testcase Type field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.TestCaseType.query.all()), 'Testcase Type field holds NOT ALL items'
    assert [opt.attrib.get('value') for opt in options if opt.attrib.get('selected')] == \
        [str(item.testcase_type.id)], 'Testcase Type field selected value is wrong'
    # hidden multi-select field
    field = form.xpath(f'.//select[@id="{sub_name}"]')
    assert bool(field), 'Hidden TestStepSequence select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_sub_items_set, \
        'Hidden TestStepSequence select field holds NOT ALL items'
    # visible select field
    assert bool(form.xpath(f'.//input[@data-name="{sub_name}"]')), 'TestStepSequence field absents'
    # chip area
    chip_area = form.xpath(f'.//div[@id="chips_{sub_name}"]')
    assert bool(chip_area), 'TestStepSequence chip area absents'
    chips = chip_area[0].xpath('./div')
    assert set(chip.attrib.get('data-id') for chip in chips) == sub_items_set
    # datalist
    datalist = form.xpath(f'.//datalist[@id="{sub_name}Opt"]')
    assert bool(datalist), 'TestStepSequence data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_sub_items_set), 'TestStepSequence datalist holds NOT ALL items'
    assert set(opt.attrib.get('data-id') for opt in options if opt.attrib.get('disabled')) == sub_items_set
    
    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_post_edit_testcase(client, test_user, populate):
    # get random testcase item
    item = choice(models.TestCase.query.all())
    url = f'/testcase/{item.uuid}/edit'

    # login
    login(client, test_user)
    update_data = {
        'name': 'Updated TestCase',
        'description': 'Updated TestCase Description',
        'classname': 1,
        'browser_type': 1,
        'testcase_type': 1,
    }

    # post update without CSRF
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestCase WITHOUT CSRF. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestCase WITHOUT CSRF. Edit Form absents'
    assert not bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake classname items
    update_data['csrf_token'] = get_csrf(client, url)
    update_data['classname'] = 0
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestCase with FAKE classname. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestCase with FAKE classname. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE classname'

    # post fake browser type items
    update_data['classname'] = 1
    update_data['browser_type'] = 0
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestCase with FAKE browser type. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestCase with FAKE browser type. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE browser type'

    # post fake testcase type items
    update_data['browser_type'] = 1
    update_data['testcase_type'] = 0
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestCase with FAKE testcase type. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestCase with FAKE testcase type. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE testcase type'

    # post fake subitems
    update_data['testcase_type'] = 1
    update_data['teststep_sequences'] = [str(uuid.uuid4())]
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestCase with FAKE sub items. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestCase with FAKE sub items. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE sub items'

    # update item
    update_data['teststep_sequences'] = [x.uuid for x in sample(models.TestStepSequence.query.all(), k=populate[1])]
    r = client.post(url, data=update_data)
    assert r.status_code == 302, f'Post Edit TestCase with CORRECT form. Status code: {r.status_code}'

    # assert updated element
    item = models.TestCase.query.get(item.id)
    assert item.name == update_data.get('name')
    assert item.description == update_data.get('description')
    assert item.classname.id == update_data.get('classname')
    assert item.browser_type.id == update_data.get('browser_type')
    assert item.testcase_type.id == update_data.get('testcase_type')
    assert set(x.uuid for x in item.teststep_sequences) == set(update_data.get('teststep_sequences'))

#
# TestStepSequence Forms
#

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_get_create_teststep_sequence(client, test_user, populate, default_supports):
    # get random teststep sequence item
    url = f'/teststep_sequence/new'
    sub_name = 'teststeps'
    avail_sub_items_set = set(i.uuid for i in models.TestStepExecution.query.filter_by(teststep_sequence_id=None).all())

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'GET Create Form. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)
    form = page.xpath('//form[@id="create-form"]')
    assert len(form) == 1, f'Create Form absents'
    form = form[0]

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Name field value is NOT empty"
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value,  "Description field value is NOT empty"
    # classname
    field = form.xpath('.//select[@id="classname"]')
    assert bool(field), 'ClassName field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'ClassName field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.ClassName.query.all()), 'ClassName field holds NOT ALL items'
    assert [opt.text for opt in options if opt.attrib.get('selected')] == \
        [default_supports.get('CLASSNAME_TESTSTEP')], 'Classname field default value is wrong'
    # hidden multi-select field
    field = form.xpath(f'.//select[@id="{sub_name}"]')
    assert bool(field), 'Hidden TestStep select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_sub_items_set, \
        'Hidden TestStep select field holds NOT ALL items'
    # visible select field
    assert bool(form.xpath(f'.//input[@data-name="{sub_name}"]')), 'TestStep field absents'
    # chip area
    chip_area = form.xpath(f'.//div[@id="chips_{sub_name}"]')
    assert bool(chip_area), 'TestStep chip area absents'
    assert not bool(chip_area[0].text.strip()), 'Chip area is not empty'    # datalist
    datalist = form.xpath(f'.//datalist[@id="{sub_name}Opt"]')
    assert bool(datalist), 'TestStep data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_sub_items_set), 'TestStep datalist holds NOT ALL items'
    assert set(opt.attrib.get('data-id') for opt in options) == avail_sub_items_set, 'Datalist holds NOT ALL items'
    
    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_post_create_teststep_sequence(client, test_user, populate):
    # login
    url = f'/teststep_sequence/new'
    login(client, test_user)

    # post empty form
    r = client.post(url, data={})
    assert r.status_code == 200, f'Post empty Create Form. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Create Form absents'
    assert len(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')) == 3, 'No error message for EMPTY form'

    item_data = {
        'name': 'New Test TestStepSequence',
        'description': 'New Test TestStepSequence Description',
        'classname': 1,
    }

    # post create without CSRF
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form  WITHOUT CSRF. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form WITHOUT CSRF. Create Form absents'
    assert not bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake classname items
    item_data['csrf_token'] = get_csrf(client, url)
    item_data['classname'] = 0
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE classname. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form with FAKE classname. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE classname'

    # post fake sub items
    item_data['classname'] = choice(models.ClassName.query.all()).id
    item_data['teststeps'] = [str(uuid.uuid4())]
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE sub items. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form with FAKE sub items. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE sub items'

    # create item
    item_data['teststeps'] = [x.uuid for x in sample(
        models.TestStepExecution.query.filter_by(teststep_sequence_id=None).all(),
        k=2 if populate[0] == 1 else 0,
    )]
    r = client.post(url, data=item_data)
    assert r.status_code == 302, f'Post Create Form with CORRECT form. Status code: {r.status_code}'

    # assert updated element
    item = models.TestStepSequence.query.filter_by(name=item_data.get('name')).first()
    assert bool(item), 'Item NOT saved to DB'
    assert item.description == item_data.get('description')
    assert item.classname.id == item_data.get('classname')
    assert set(x.uuid for x in item.teststeps) == set(item_data.get('teststeps'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_get_edit_teststep_sequence(client, test_user, populate):
    # get random teststep sequence item
    item = choice(models.TestStepSequence.query.all())
    url = f'/teststep_sequence/{item.uuid}/edit'
    sub_items_set = set(i.uuid for i in item.teststeps)
    sub_name = 'teststeps'
    avail_sub_items_set = set(i.uuid for i in models.TestStepExecution.query.filter(
        (models.TestStepExecution.teststep_sequence_id==item.id) | (models.TestStepExecution.teststep_sequence_id==None)).all())

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'GET Edit TestCaseSequence. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="edit-form"]')
    assert len(form) == 1, f'Edit Form absents'
    form = form[0]

    # reload item
    item = models.TestStepSequence.query.get(item.id)

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = field[0].attrib.get('value')
    assert value == item.name, f"Name field value '{value}' does NOT match item name '{item.name}'"
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = field[0].attrib.get('value')
    assert value == item.description, f"Description field value '{value}' does NOT match item description '{item.description}'"
    # classname
    field = form.xpath('.//select[@id="classname"]')
    assert bool(field), 'ClassName field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'ClassName field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.ClassName.query.all()), 'ClassName field holds NOT ALL items'
    assert [opt.attrib.get('value') for opt in options if opt.attrib.get('selected')] == \
        [str(item.classname.id)], 'Classname field selected value is wrong'
    # hidden multi-select field
    field = form.xpath(f'.//select[@id="{sub_name}"]')
    assert bool(field), 'Hidden TestStep select field absents'
    assert set(opt.attrib.get('value') for opt in field[0].xpath('./option')) == avail_sub_items_set, \
        'Hidden TestStep select field holds NOT ALL items'
    # visible select field
    assert bool(form.xpath(f'.//input[@data-name="{sub_name}"]')), 'TestStep field absents'
    # chip area
    chip_area = form.xpath(f'.//div[@id="chips_{sub_name}"]')
    assert bool(chip_area), 'TestStep chip area absents'
    chips = chip_area[0].xpath('./div')
    assert set(chip.attrib.get('data-id') for chip in chips) == sub_items_set
    # datalist
    datalist = form.xpath(f'.//datalist[@id="{sub_name}Opt"]')
    assert bool(datalist), 'TestStep data absents'
    options = datalist[0].xpath('./option')
    assert len(options) == len(avail_sub_items_set), 'TestStep datalist holds NOT ALL items'
    assert set(opt.attrib.get('data-id') for opt in options if opt.attrib.get('disabled')) == sub_items_set
    
    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_post_edit_teststep_sequence(client, test_user, populate):
    # get random teststep sequence item
    item = choice(models.TestStepSequence.query.all())
    url = f'/teststep_sequence/{item.uuid}/edit'

    # login
    login(client, test_user)
    update_data = {
        'name': 'Updated TestStepSequence',
        'description': 'Updated TestStepSequence Description',
        'classname': 1,
    }

    # post update without CSRF
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestStepSequence WITHOUT CSRF. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestStepSequence WITHOUT CSRF. Edit Form absents'
    assert not bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake classname items
    update_data['csrf_token'] = get_csrf(client, url)
    update_data['classname'] = 0
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestStepSequence with FAKE classname. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestStepSequence with FAKE classname. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE classname'

    # post fake sub items
    update_data['classname'] = 1
    update_data['teststeps'] = [str(uuid.uuid4())]
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestStepSequence with FAKE sub items. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestStepSequence with FAKE sub items. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE sub items'

    # update item
    update_data['teststeps'] = [models.TestStepExecution.query.filter_by(teststep_sequence_id=item.id).first().uuid]
    r = client.post(url, data=update_data)
    assert r.status_code == 302, f'Post Edit TestStepSequence with CORRECT form. Status code: {r.status_code}'

    # assert updated element
    item = models.TestStepSequence.query.get(item.id)
    assert item.name == update_data.get('name')
    assert item.description == update_data.get('description')
    assert item.classname.id == update_data.get('classname')
    assert set(x.uuid for x in item.teststeps) == set(update_data.get('teststeps'))


#
# TestStep Forms
#
@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_get_create_teststep(client, test_user, populate, default_supports, teststep_selects):
    # set url
    url = f'/teststep/new'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'GET Create Form. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="create-form"]')
    assert len(form) == 1, f'Create Form absents'
    form = form[0]

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Name field value is NOT empty'"
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Description field value is NOT empty"
    # activity type
    field = form.xpath('.//select[@id="activity_type"]')
    assert bool(field), 'ActivityType field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'ActivityType field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.ActivityType.query.all()), 'Activity field holds NOT ALL items'
    assert [opt.text for opt in options if opt.attrib.get('selected')] == \
        [default_supports.get('ACTIVITY_TYPE')], 'ActivityType field default value is wrong'
    # locator type
    field = form.xpath('.//select[@id="locator_type"]')
    assert bool(field), 'LocatorType field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'LocatorType field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set([(f'{item.id}', item.name) for item in models.LocatorType.query.all()] + [('0', 'none')]), 'LocatorType field holds NOT ALL items'
    assert [opt.text for opt in options if opt.attrib.get('selected')] == ['none'], 'LocatorType field default value is wrong'
    # locator
    field = form.xpath('.//input[@id="locator"]')
    assert bool(field), 'Locator field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Locator field value is NOT empty"
    # value
    field = form.xpath('.//input[@id="value"]')
    assert bool(field), 'Value field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Value field value is NOT empty"
    # value2
    field = form.xpath('.//input[@id="value2"]')
    assert bool(field), 'Value2 field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Value2 field value is NOT empty"
    # comparison
    field = form.xpath('.//select[@id="comparison"]')
    assert bool(field), 'Comparison field absents'
    options = field[0].xpath('./option')
    required_options = set(teststep_selects.get('comparison'))
    assert bool(options), 'Comparison field does NOT hold options'
    assert set(opt.text for opt in options) == required_options
    assert [opt.text for opt in options if opt.attrib.get('selected')] == ['none'], 'Comparison field default value is wrong'
    # optional
    field = form.xpath('.//select[@id="optional"]')
    assert bool(field), 'Optional field absents'
    options = field[0].xpath('./option')
    required_options = set(teststep_selects.get('optional'))
    assert bool(options), 'Optional field does NOT hold options'
    assert set(opt.text for opt in options) == required_options
    assert [opt.text for opt in options if opt.attrib.get('selected')] == ['False'], 'Optional field default value is wrong'
    # timeout
    field = form.xpath('.//input[@id="timeout"]')
    assert bool(field), 'Timeout field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Timeout field value is NOT empty"
    # release
    field = form.xpath('.//input[@id="release"]')
    assert bool(field), 'Release field absents'
    value = bool(field[0].attrib.get('value'))
    assert not value, "Release field value is NOT empty"

    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_post_create_teststep(client, test_user, populate, teststep_selects):
    # login
    url = f'/teststep/new'
    login(client, test_user)

    # post empty form
    r = client.post(url, data={})
    assert r.status_code == 200, f'Post empty Create Form. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Create Form absents'
    assert len(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')) == 6, 'No error message for EMPTY form'

    item_data = {
        'name': 'New Test TestStep (required filds only)',
        'description': 'New Test TestStep Description',
        'activity_type': 1,
        'locator_type': 0,
        'comparison': 0,
        'optional': 0,
    }

    # post update without CSRF
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form WITHOUT CSRF. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form WITHOUT CSRF. Create Form absents'
    assert not bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake activity item
    item_data['csrf_token'] = get_csrf(client, url)
    item_data['activity_type'] = 0
    r = client.post(url, data=item_data)
    assert r.status_code == 200, f'Post Create Form with FAKE activity. Status code: {r.status_code}'
    # assert create page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="create-form"]')), f'Post Create Form with FAKE activity. Create Form absents'
    assert bool(page.xpath('//form[@id="create-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE activity'

    # create item with required fields
    item_data['activity_type'] = 1
    r = client.post(url, data=item_data)
    assert r.status_code == 302, f'Post Create Form with CORRECT form (required data only). Status code: {r.status_code}'

    # assert created element
    item = models.TestStepExecution.query.filter_by(name=item_data.get('name')).first()
    assert bool(item.name), 'Item with required only data was NOT saved to DB' 
    assert item.description == item_data.get('description')
    assert item.activity_type.id == item_data.get('activity_type')
    assert item.locator_type is None
    assert item.locator is None
    assert item.value is None
    assert item.value2 is None
    assert item.comparison is None
    assert not item.optional
    assert item.timeout is None
    assert item.release is None

    # create item with all fields filled
    item_data.update({
        'name': 'New Test TestStep (fully filled)',
        'locator_type': choice(models.LocatorType.query.all()).id,
        'locator': 'Test Locator',
        'value': 'Test Value',
        'value2': 'Test Value-2',
        'comparison': randrange(1, len(teststep_selects.get("comparison"))),
        'optional': 1,
        'timeout': '1000.7',
        'release': 'Test Release',
    }) 
    r = client.post(url, data=item_data)
    assert r.status_code == 302, f'Post Create Form with CORRECT form (fully filled). Status code: {r.status_code}'

    # assert updated element
    item = models.TestStepExecution.query.filter_by(name=item_data.get('name')).first()
    assert bool(item.name), 'Item with fully filled data was NOT saved to DB' 
    assert item.locator_type.id == item_data.get('locator_type')
    assert item.locator == item_data.get('locator')
    assert item.value == item_data.get('value')
    assert item.value2 == item_data.get('value2')
    assert item.comparison == teststep_selects.get("comparison")[item_data.get('comparison')]
    assert item.optional
    assert str(item.timeout) == item_data.get('timeout')
    assert item.release == item_data.get('release')

@pytest.mark.parametrize('count', [(1,3), (5,1)])
def test_get_edit_teststep(client, test_user, populate, teststep_selects):
    # get random teststep item
    item = choice(models.TestStepExecution.query.all())
    url = f'/teststep/{item.uuid}/edit'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200, f'GET Edit TestCaseSequence. Status code: {r.status_code}'

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="edit-form"]')
    assert len(form) == 1, f'Edit Form absents'
    form = form[0]

    # reload item
    item = models.TestStepExecution.query.get(item.id)

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field absents'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # name
    field = form.xpath('.//input[@id="name"]')
    assert bool(field), 'Name field absents'
    value = field[0].attrib.get('value')
    assert value == item.name, f"Name field value '{value}' does NOT match item name '{item.name}'"
    # description
    field = form.xpath('.//textarea[@id="description"]')
    assert bool(field), 'Description field absents'
    value = field[0].attrib.get('value')
    assert value == item.description, f"Description field value '{value}' does NOT match item description '{item.description}'"
    # activity type
    field = form.xpath('.//select[@id="activity_type"]')
    assert bool(field), 'ActivityType field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'ActivityType field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set((f'{item.id}', item.name) for item in models.ActivityType.query.all()), 'Activity field holds NOT ALL items'
    assert [opt.attrib.get('value') for opt in options if opt.attrib.get('selected')] == \
        [str(item.activity_type.id)], 'ActivityType field selected value is wrong'
    # locator type
    field = form.xpath('.//select[@id="locator_type"]')
    assert bool(field), 'LocatorType field absents'
    options = field[0].xpath('./option')
    assert bool(options), 'LocatorType field does NOT hold options'
    assert set((opt.attrib.get('value'), opt.text) for opt in options) == \
        set([(f'{item.id}', item.name) for item in models.LocatorType.query.all()] + [('0', 'none')]), 'LocatorType field holds NOT ALL items'
    assert [opt.attrib.get('value') for opt in options if opt.attrib.get('selected')] == \
        [str(item.locator_type.id) if item.locator_type else '0'], 'LocatorType field selected value is wrong'
    # locator
    field = form.xpath('.//input[@id="locator"]')
    assert bool(field), 'Locator field absents'
    value = field[0].attrib.get('value') or None
    assert value == item.locator, f"Locator field value '{value}' does NOT match item locator '{item.locator}'"
    # value
    field = form.xpath('.//input[@id="value"]')
    assert bool(field), 'Value field absents'
    value = field[0].attrib.get('value') or None
    assert value == item.value, f"Value field value '{value}' does NOT match item locator '{item.value}'"
    # value2
    field = form.xpath('.//input[@id="value2"]')
    assert bool(field), 'Value2 field absents'
    value = field[0].attrib.get('value') or None
    assert value == item.value2, f"Value2 field value '{value}' does NOT match item locator '{item.value2}'"
    # comparison
    field = form.xpath('.//select[@id="comparison"]')
    assert bool(field), 'Comparison field absents'
    options = field[0].xpath('./option')
    required_options = set(teststep_selects.get('comparison'))
    assert bool(options), 'Comparison field does NOT hold options'
    assert set(opt.text for opt in options) == required_options
    assert [opt.text for opt in options if opt.attrib.get('selected')] == [str(item.comparison).lower()], \
        'Comparison field value does NOT match'
    # optional
    field = form.xpath('.//select[@id="optional"]')
    assert bool(field), 'Optional field absents'
    options = field[0].xpath('./option')
    required_options = set(teststep_selects.get('optional'))
    assert bool(options), 'Optional field does NOT hold options'
    assert set(opt.text for opt in options) == required_options
    assert ([opt.text for opt in options if opt.attrib.get('selected')] == ['True']) == bool(item.comparison), \
        'Optional field value does NOT match'
    # timeout
    field = form.xpath('.//input[@id="timeout"]')
    assert bool(field), 'Timeout field absents'
    value = field[0].attrib.get('value') or 'None'
    assert value == str(item.timeout), f"Timeout field value '{value}' does NOT match item timeout '{item.timeout}'"
    # release
    field = form.xpath('.//input[@id="release"]')
    assert bool(field), 'Release field absents'
    value = field[0].attrib.get('value') or None
    assert value == item.release, f"Release field value '{value}' does NOT match item locator '{item.release}'"

    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))


@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_post_edit_teststep(client, test_user, populate, teststep_selects):
    # get random teststep item
    item = choice(models.TestStepExecution.query.all())
    url = f'/teststep/{item.uuid}/edit'

    # login
    login(client, test_user)
    update_data = {
        'name': 'Updated TestStep',
        'description': 'Updated TestStep Description',
        'activity_type': 1,
        'locator_type': 0,
        'comparison': 0,
        'optional': 0,
    }

    # post update without CSRF
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestStepSequence WITHOUT CSRF. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestStepSequence WITHOUT CSRF. Edit Form absents'
    assert not bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), \
        'There SHOULD be NO ERROR MESSAGE posting WITHOUT CSRF'

    # post fake activity item
    update_data['csrf_token'] = get_csrf(client, url)
    update_data['activity_type'] = 0
    r = client.post(url, data=update_data)
    assert r.status_code == 200, f'Post Edit TestStep with FAKE activity. Status code: {r.status_code}'
    # assert edit page
    page = html.fromstring(r.data)
    assert bool(page.xpath('//form[@id="edit-form"]')), f'Post Edit TestStep with FAKE activity. Edit Form absents'
    assert bool(page.xpath('//form[@id="edit-form"]/p[contains(@class, "form-error")]')), 'No error message for FAKE activity'

    # update item
    update_data.update({
        'activity_type': 1,
        'locator_type': 0,
        'locator': 'updated locator',
        'value': 'updated value',
        'comparison': 0,
        'optional': 0,
        'timeout': '1000.7',
        'release': 'updated release',
    }) 
    r = client.post(url, data=update_data)
    assert r.status_code == 302, f'Post Edit TestStepSequence with CORRECT form. Status code: {r.status_code}'

    # assert updated element
    item = models.TestStepExecution.query.get(item.id)
    assert item.name == update_data.get('name')
    assert item.description == update_data.get('description')
    assert item.activity_type.id == update_data.get('activity_type')
    assert item.locator_type is None
    assert item.locator == update_data.get('locator')
    assert item.value == update_data.get('value')
    assert item.value2 is None
    assert item.comparison is None
    assert not item.optional
    assert str(item.timeout) == update_data.get('timeout')
    assert item.release == update_data.get('release')

    # update items not required selects
    update_data['locator_type'] = choice(models.LocatorType.query.all()).id
    update_data['comparison'] = randrange(1, len(teststep_selects.get("comparison")))
    update_data['optional'] = 1
    r = client.post(url, data=update_data)
    assert r.status_code == 302, f'Post Edit TestStepSequence with CORRECT form. Status code: {r.status_code}'

    # assert updated element
    item = models.TestStepExecution.query.get(item.id)
    assert item.locator_type.id == update_data.get('locator_type')
    assert item.comparison == teststep_selects.get("comparison")[update_data.get('comparison')]
    assert item.optional


#
# Fake Item
#
def test_create_fake_item(client, test_user):
    # set-up
    fake_item = 'fakeitem'
    url = f'/{fake_item}/new'

    # authorized access
    login(client, test_user)

    # GET method
    r = client.get(url, follow_redirects=True)
    assert r.status_code == 200, f'Edit fake item. Status code: {r.status_code}'
    # assert warning Messagese
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-warning")]')
    assert bool(msg), 'Warning message is absent'
    assert fake_item in msg[0].text, 'Warning message does NOT mention fake API host'

    # POST method
    r = client.post(url, follow_redirects=True)
    assert r.status_code == 200, f'Edit fake item. Status code: {r.status_code}'
    # assert warning Messagese
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-warning")]')
    assert bool(msg), 'Warning message is absent'
    assert fake_item in msg[0].text, 'Warning message does NOT mention fake API host'

def test_edit_fake_item(client, test_user):
    # set-up
    fake_item_id = uuid.uuid4()
    fake_item = 'fakeitem'
    url = f'/{fake_item}/{fake_item_id}/edit'

    # authorized access
    login(client, test_user)

    # GET method
    r = client.get(url, follow_redirects=True)
    assert r.status_code == 200, f'Edit fake item. Status code: {r.status_code}'
    # assert warning Messagese
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-warning")]')
    assert bool(msg), 'Warning message is absent'
    assert fake_item in msg[0].text, 'Warning message does NOT mention fake API host'

    # POST method
    r = client.post(url, follow_redirects=True)
    assert r.status_code == 200, f'Edit fake item. Status code: {r.status_code}'
    # assert warning Messagese
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-warning")]')
    assert bool(msg), 'Warning message is absent'
    assert fake_item in msg[0].text, 'Warning message does NOT mention fake API host'


