##########################################
#                                        #
#    ENSURE api service is up		     #
#                                        #
##########################################

import pytest
import os
import re
from lxml import html
from app import models
from tests.supports import login, path_to_data, get_csrf
import uuid


@pytest.mark.parametrize('count', [(1,1)])
def test_teststep(client, test_user, populate):
	# set-up
    testrun = models.Testrun.query.first()
    url = f'/testrun/{testrun.uuid}/run'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    
    # assert response
    assert r.status_code == 200
    page = html.fromstring(r.data)
    execution_status = page.xpath('//div[@id="execution-status"]')
    assert bool(execution_status), 'Card with Testrun Execution Status is absent'
    summary_link = execution_status[0].xpath('./a')
    assert bool(summary_link), 'Call Summary link is absent'

    # assert call id
    match_call_id = re.search(r'[\w-]+$', summary_link[0].attrib.get('href'))
    assert bool(match_call_id), 'Call Summary link does not contains Call ID'
    call_id = match_call_id.group()
    tr_call = models.TestRunCall.query.filter_by(testrun_id=testrun.id).first()
    assert tr_call, 'Testrun has no call'
    assert str(tr_call) == call_id, 'Call ID does not match'


def test_execute_fake_testrun(client, test_user):
    # execute url for fake testrun
    fake_tr_id = uuid.uuid4()
    url = f'/testrun/{fake_tr_id}/run'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.get(url, follow_redirects=True)
    # assert error message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-danger")]')  
    assert bool(msg), 'Error message is absent'
    assert str(fake_tr_id) in msg[0].text, 'Error message does NOT mention fake Testrun ID'


@pytest.mark.parametrize('count', [(1,1)])
def test_execute_uncomplete_testrun(client, test_user, populate):
    # set-up
    testrun = models.Testrun.query.first()
    urls = {
        'run': f'/testrun/{testrun.uuid}/run',
        'delete': f'/testcase_sequence/{testrun.testcase_sequences[0].uuid}/delete',
    }

    # delete testcase sequence of the testrun
    login(client, test_user)
    r = client.delete(urls.get('delete'))

    r = client.get(urls.get('run'), follow_redirects=True)
    # assert error Messagese
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-danger")]')
    assert bool(msg), 'Error message is absent'
    assert testrun.uuid in msg[0].text, 'Error message does NOT mention fake Testrun ID'


@pytest.mark.parametrize('count', [(1,1)])
def test_execute_fake_api_host(application, test_user, populate):
    # set-up
    testrun = models.Testrun.query.first()
    url = f'/testrun/{testrun.uuid}/run'
    client = application.test_client()
    application.config['BAANGT_API_HOST'] = 'fake.host'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.get(url, follow_redirects=True)
    # assert error Messagese
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-danger")]')
    assert bool(msg), 'Error message is absent'
    assert application.config['BAANGT_API_HOST'] in msg[0].text, 'Error message does NOT mention fake API host'

