import pytest
from tests.supports import login, path_to_data, remove_file, TRTestHelper
from app import models
from lxml import html
from random import choice
import os
import json
import uuid
import xlrd


#
# Export Testrun
#
def test_export_errors(client, test_user):
    # login
    login(client, test_user)

    # get not existing item
    item_id = str(uuid.uuid4())
    url = f'/testrun/{item_id}/export'
    
    # xlsx
    r = client.get(f'{url}/xlsx')
    assert r.status_code == 400
    assert b'error' in r.data
    assert item_id.encode() in r.data
    
    # json
    r = client.get(f'{url}/json')
    assert r.status_code == 400
    assert b'error' in r.data
    assert item_id.encode() in r.data
    
    # wrong format
    wrong_format = 'pdf' 
    r = client.get(f'{url}/{wrong_format}')
    assert r.status_code == 400
    assert b'error' in r.data
    assert wrong_format.encode() in r.data

@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_export_json(client, test_user, populate):
    # get random testrun
    item = choice(models.Testrun.query.all())
    url = f'/testrun/{item.uuid}/export/json'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200

    # assert response
    response_json = json.loads(r.data)
    assert 'success' in response_json
    assert all(not response_json.get(key) is None for key in ['filename', 'url'])

    # assert exported file
    r = client.get(response_json.get('url'))
    assert r.status_code == 200

    exported_item = json.loads(r.data)
    # reload item
    item = models.Testrun.query.get(item.id)
    # common data
    assert set(exported_item.keys()) == {'name', 'testrun', 'globals'}, 'Invalid keys in exported object' 
    assert exported_item['name'] == item.name, 'Testrun name does NOT match'
    assert set(exported_item['testrun'].keys()) == {'GC.KWARGS_TESTRUNATTRIBUTES'}, 'Invalid key in exported Testrun'
    assert set(exported_item['testrun']['GC.KWARGS_TESTRUNATTRIBUTES'].keys()) == {'GC.EXPORT_FORMAT', 'GC.STRUCTURE_TESTCASESEQUENCE'}, \
        'Invalid keys in exported Testrun attributes'
    assert set(exported_item['testrun']['GC.KWARGS_TESTRUNATTRIBUTES']['GC.EXPORT_FORMAT'].keys()) == \
        {'GC.EXPORT_FORMAT', 'GC.EXP_FIELDLIST'}, 'Invalid keys in exported Testrun format'
    assert exported_item['globals'] == (json.loads(item.global_settings) if item.global_settings else None), 'Global Settings does not match' 
    
    # testcase sequences
    exported_item = exported_item['testrun']['GC.KWARGS_TESTRUNATTRIBUTES']['GC.STRUCTURE_TESTCASESEQUENCE']
    assert len(exported_item) == len(item.testcase_sequences), 'Invalid number of exported TestCaseSequences'
    for tcs_index, exported_tcs in exported_item.items():
        tcs = item.testcase_sequences[int(tcs_index)-1]
        assert len(exported_tcs) == 2, f'TestCaseSequence {tcs_index}: invalid number of attributes'
        assert exported_tcs[0] == tcs.classname.name, f'TestCaseSequence {tcs_index}: ClassName does NOT match'
        assert exported_tcs[1].get('SequenceClass') == tcs.classname.name, f'TestCaseSequence {tcs_index}: ClassName does NOT match'
        assert exported_tcs[1].get('GC.DATABASE_FILENAME') == tcs.datafiles[0].uuid, \
            f'TestCaseSequence {tcs_index}: DataFile name does NOT match' 
        assert exported_tcs[1].get('GC.DATABASE_SHEETNAME') == tcs.datafiles[0].sheet, \
            f'TestCaseSequence {tcs_index}: DataFile sheet does NOT match' 
        assert exported_tcs[1].get('GC.EXECUTION_PARALLEL') == 1, f'TestCaseSequence {tcs_index}: wrong EXECUTION_PARALLEL value'
        assert exported_tcs[1].get('GC.DATABASE_LINES') == '0-999999', f'TestCaseSequence {tcs_index}: wrong DATABASE_LINES value'
        
        # testcases
        assert 'GC.STRUCTURE_TESTCASE' in exported_tcs[1], f'TestCaseSequence {tcs_index}: TestCases missed'
        assert len(exported_tcs[1].get('GC.STRUCTURE_TESTCASE')) == len(tcs.testcases), \
            f'TestCaseSequence {tcs_index}: Invalid number of exported TestCases'
        for tc_index, exported_tc in exported_tcs[1]['GC.STRUCTURE_TESTCASE'].items():
            tc = tcs.testcases[int(tc_index)-1]
            assert len(exported_tc) == 3, f'TestCase {tcs_index}-{tc_index}: invalid number of attributes'
            assert exported_tc[0] == tc.classname.name, f'TestCase {tcs_index}-{tc_index}: ClassName does NOT match'
            assert exported_tc[1].get('GC.KWARGS_TESTCASETYPE') == tc.testcase_type.name, \
                f'TestCase {tcs_index}-{tc_index}: TestCaseType does NOT match'
            assert exported_tc[1].get('GC.KWARGS_BROWSER') == tc.browser_type.name, \
                f'TestCase {tcs_index}-{tc_index}: BrowserType does NOT match'
            assert exported_tc[1].get('GC.BROWSER_ATTRIBUTES') == '', f'TestCase {tcs_index}-{tc_index}: Browser Attributes NOT empty'

            # teststep sequences
            assert 'GC.STRUCTURE_TESTSTEP' in exported_tc[2], f'TestCase {tcs_index}-{tc_index}: TestStepSequences missed'
            assert len(exported_tc[2].get('GC.STRUCTURE_TESTSTEP')) == len(tc.teststep_sequences), \
                f'TestCase {tcs_index}-{tc_index}: Invalid number of exported TestStepSequences'
            for tss_index, exported_tss in exported_tc[2]['GC.STRUCTURE_TESTSTEP'].items():
                tss = tc.teststep_sequences[int(tss_index)-1]
                assert len(exported_tss) == 2, f'TestStepSequence {tcs_index}-{tc_index}-{tss_index}: invalid number of attributes'
                assert exported_tss[0].get('TestStepClass') == tss.classname.name, \
                    f'TestStepSequence {tcs_index}-{tc_index}-{tss_index}: ClassName does NOT match'

                # teststeps
                assert 'GC.STRUCTURE_TESTSTEPEXECUTION' in exported_tss[1], \
                    f'TestStepSequence {tcs_index}-{tc_index}-{tss_index}: TestSteps missed'
                assert len(exported_tss[1].get('GC.STRUCTURE_TESTSTEPEXECUTION')) == len(tss.teststeps), \
                    f'TestCase {tcs_index}-{tc_index}: Invalid number of exported TestSteps'
                for ts_index, exported_ts in exported_tss[1]['GC.STRUCTURE_TESTSTEPEXECUTION'].items():
                    ts = tss.teststeps[int(ts_index)-1]
                    assert exported_ts.get('Activity') == ts.activity_type.name, \
                        f'TestStep {tcs_index}-{tc_index}-{tss_index}-{ts_index}: ActivityType does NOT match'
                    assert exported_ts.get('LocatorType') == str(ts.locator_type or ''), \
                        f'TestStep {tcs_index}-{tc_index}-{tss_index}-{ts_index}: LocatorType does NOT match'
                    assert exported_ts.get('Locator') == (ts.locator or '')
                    assert exported_ts.get('Value') == (ts.value or '')
                    assert exported_ts.get('Comparison') == (ts.comparison or '')
                    assert exported_ts.get('Value2') == (ts.value2 or '')
                    assert exported_ts.get('Timeout') == (ts.timeout or '')
                    assert exported_ts.get('Optional') == (ts.optional or '')
                    assert exported_ts.get('Release') == (ts.release or '')

    # delete exported file
    remove_file(os.path.join(os.getcwd(), 'app', response_json.get('url')[1:]))


@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_export_xlsx(client, test_user, populate):
    # get random testrun
    item = choice(models.Testrun.query.all())
    url = f'/testrun/{item.uuid}/export/xlsx'

    # unauthorized access
    r = client.get(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 200

    # assert response
    response_json = json.loads(r.data)
    assert 'success' in response_json
    assert all(not response_json.get(key) is None for key in ['filename', 'url'])

    # assert exported file
    r = client.get(response_json.get('url'))
    assert r.status_code == 200
    exported_item = xlrd.open_workbook(file_contents=r.data)
    
    # reload item
    item = models.Testrun.query.get(item.id)
    tr_helper = TRTestHelper(item)
    with open(os.path.join(path_to_data, 'xl_definition.json'), 'r') as f:
        xl_structure = json.load(f)
    # sheet names
    sheet_names = list(xl_structure.keys())
    assert set(exported_item.sheet_names()) == set(sheet_names), 'Sheets do not match'
   
    # sheet TestRun
    cur_sheet_name = sheet_names[0]
    sheet = exported_item.sheet_by_name(cur_sheet_name)
    assert sheet.nrows == 2, f'Sheet "{cur_sheet_name}": number of rows does not match'
    assert sheet.ncols == 2, f'Sheet "{cur_sheet_name}": number of cols does not match'
    for col, (header, value) in enumerate(xl_structure[cur_sheet_name].items()):
        assert sheet.cell(0, col).value == header
        assert sheet.cell(1, col).value == value
    
    # sheet TestCaseSequence
    cur_sheet_name = sheet_names[1]
    sheet = exported_item.sheet_by_name(cur_sheet_name)
    col_indexes = {h: index for index, h in enumerate(xl_structure[cur_sheet_name])}
    assert sheet.nrows == tr_helper.get_num(TRTestHelper.TCS)+1, f'Sheet "{cur_sheet_name}": number of rows does not match'
    assert sheet.ncols == len(xl_structure[cur_sheet_name]), f'Sheet "{cur_sheet_name}": number of cols does not match'
    for tcs_index, tcs in tr_helper.get_indexes(TRTestHelper.TCS):
        assert int(sheet.cell(tcs_index, col_indexes.get('Number')).value) == tcs_index
        assert sheet.cell(tcs_index, col_indexes.get('SequenceClass')).value == tcs.classname.name
        assert sheet.cell(tcs_index, col_indexes.get('TestDataFileName')).value == tcs.datafiles[0].filename
        assert sheet.cell(tcs_index, col_indexes.get('Sheetname')).value == tcs.datafiles[0].sheet
    
    # sheet TestCase
    cur_sheet_name = sheet_names[2]
    sheet = exported_item.sheet_by_name(cur_sheet_name)
    col_indexes = {h: index for index, h in enumerate(xl_structure[cur_sheet_name])}
    assert sheet.nrows == tr_helper.get_num(TRTestHelper.TC)+1, f'Sheet "{cur_sheet_name}": number of rows does not match'
    assert sheet.ncols == len(xl_structure[cur_sheet_name]), f'Sheet "{cur_sheet_name}": number of cols does not match'
    for tcs_index, tc_index, tc in tr_helper.get_indexes(TRTestHelper.TC):
        assert int(sheet.cell(tc_index, col_indexes.get('TestCaseSequenceNumber')).value) == tcs_index
        assert int(sheet.cell(tc_index, col_indexes.get('TestCaseNumber')).value) == tc_index
        assert sheet.cell(tc_index, col_indexes.get('TestCaseClass')).value == tc.classname.name
        assert sheet.cell(tc_index, col_indexes.get('TestCaseType')).value == tc.testcase_type.name
        assert sheet.cell(tc_index, col_indexes.get('Browser')).value == tc.browser_type.name

    # sheet TestStep
    cur_sheet_name = sheet_names[3]
    sheet = exported_item.sheet_by_name(cur_sheet_name)
    col_indexes = {h: index for index, h in enumerate(xl_structure[cur_sheet_name])}
    assert sheet.nrows == tr_helper.get_num(TRTestHelper.TSS)+1, f'Sheet "{cur_sheet_name}": number of rows does not match'
    assert sheet.ncols == len(xl_structure[cur_sheet_name]), f'Sheet "{cur_sheet_name}": number of cols does not match'
    for tcs_index, tc_index, tss_index, tss in tr_helper.get_indexes(TRTestHelper.TSS):
        assert int(sheet.cell(tss_index, col_indexes.get('TestCaseSequenceNumber')).value) == tcs_index
        assert int(sheet.cell(tss_index, col_indexes.get('TestCaseNumber')).value) == tc_index
        assert int(sheet.cell(tss_index, col_indexes.get('TestStepNumber')).value) == tss_index
        assert sheet.cell(tss_index, col_indexes.get('TestStepClass')).value == tss.classname.name

    # sheet TestStepExecution
    cur_sheet_name = sheet_names[4]
    sheet = exported_item.sheet_by_name(cur_sheet_name)
    col_indexes = {h: index for index, h in enumerate(xl_structure[cur_sheet_name])}
    assert sheet.nrows == tr_helper.get_num(TRTestHelper.TS)+1, f'Sheet "{cur_sheet_name}": number of rows does not match'
    assert sheet.ncols == len(xl_structure[cur_sheet_name]), f'Sheet "{cur_sheet_name}": number of cols does not match'
    for tcs_index, tc_index, tss_index, ts_index, ts in tr_helper.get_indexes(TRTestHelper.TS):
        print(f'TCS: {tcs_index}, TC: {tc_index}, TSS {tss_index}, TS: {ts_index}')
        assert int(sheet.cell(ts_index, col_indexes.get('TestCaseSequenceNumber')).value) == tcs_index
        assert int(sheet.cell(ts_index, col_indexes.get('TestCaseNumber')).value) == tc_index
        assert int(sheet.cell(ts_index, col_indexes.get('TestStepNumber')).value) == tss_index
        assert int(sheet.cell(ts_index, col_indexes.get('TestStepExecutionNumber')).value) == ts_index
        assert sheet.cell(ts_index, col_indexes.get('Activity')).value == ts.activity_type.name
        assert sheet.cell(ts_index, col_indexes.get('LocatorType')).value == (ts.locator_type.name if ts.locator_type else '')
        assert sheet.cell(ts_index, col_indexes.get('Locator')).value == (ts.locator or '')
        assert sheet.cell(ts_index, col_indexes.get('Value')).value == (ts.value or '')
        assert sheet.cell(ts_index, col_indexes.get('Value2')).value == (ts.value2 or '')
        assert sheet.cell(ts_index, col_indexes.get('Comparison')).value == (ts.comparison or '')
        assert sheet.cell(ts_index, col_indexes.get('Timeout')).value == (ts.timeout or '')
        assert bool(sheet.cell(ts_index, col_indexes.get('Optional')).value) == ts.optional
        assert sheet.cell(ts_index, col_indexes.get('Release')).value == (ts.release or '')

    # delete exported file
    remove_file(os.path.join(os.getcwd(), 'app', response_json.get('url')[1:]))

