import pytest
from tests.supports import login, path_to_data, get_csrf
from app import models
from lxml import html
import os
import json
from random import choice


#
# Import Testrun
#
def test_import_form(client, test_user):
    # login
    url = f'/testrun'
    login(client, test_user)
    r = client.get(url)

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="import-form"]')
    assert bool(form), f'Import Form is absent'
    form = form[0]

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field is absent'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # file field
    field = form.xpath('.//input[@id="file"]')
    assert bool(field), 'File field is absent'
    value = bool(field[0].attrib.get('value'))
    assert not value, f'Name field is NOT empty'
    # 3 datafile fields
    for i in range(3):
        field = form.xpath(f'.//input[@id="dataFile{i+1}"]')
        assert bool(field), f'DataFile-{i+1} field is absent'
        value = bool(field[0].attrib.get('value'))
        assert not value, f'DataFile-{i+1} field is NOT empty'

    # assert datafile add button
    assert bool(form.xpath('.//a[@id="dataFileExpander"]')), 'DataFile Add button is absent'

    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))


def test_import_simple(client, test_user, default_supports, db_supports):
    # url
    url = '/testrun/import'
    db_supports(5)

    # unauthorized access
    r = client.post(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized GET access
    login(client, test_user)
    r = client.get(url)
    assert r.status_code == 405, f'GET Import. Status code: {r.status_code}'

    # POST without file
    file_data = { 'csrf_token': get_csrf(client, '/testrun') }
    r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import WITHOUT file. Status code: {r.status_code}'
    # assert error message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-warning")]')), 'No error message for post WITHOUT data'

    # POST without CSRF
    with open(os.path.join(path_to_data, 'testrun_simple.xlsx'), 'rb') as f:
        r = client.post(url, data={'file': f}, follow_redirects=True)
    assert r.status_code == 200, f'POST Import WITHOUT CSFR. Status code: {r.status_code}'
    # assert error message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-warning")]')), 'No error message for post WITHOUT CSFR'

    # correct import
    with open(os.path.join(path_to_data, 'testrun_simple.xlsx'), 'rb') as f:
        file_data['file'] = f
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import Testrun. Status code: {r.status_code}'
    # assert success message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-success")]')), 'Success message is absent CORRECT import'

    # assert imported testrun
    with open(os.path.join(path_to_data, 'testrun_simple.json'), 'r') as f:
        imported_testrun = json.load(f)
    # testrun
    item = models.Testrun.query.filter_by(name=imported_testrun['testrun'].get('name')).first()
    assert bool(item), 'Imported Testrun NOT saved to DB'
    assert item.description == imported_testrun['testrun'].get('description_import')
    # testcase sequence
    assert len(item.testcase_sequences) == 1
    item = item.testcase_sequences[0]
    assert item.name == imported_testrun['testcase_sequence'].get('name')
    assert item.description == imported_testrun['testcase_sequence'].get('description_import')
    assert item.classname.name == imported_testrun['testcase_sequence'].get('classname')
    # testcase
    assert len(item.testcases) == 1
    item = item.testcases[0]
    assert item.name == imported_testrun['testcase'].get('name')
    assert item.description == imported_testrun['testcase'].get('description_import')
    assert item.classname.name == imported_testrun['testcase'].get('classname')
    assert item.browser_type.name == imported_testrun['testcase'].get('browser_type')
    assert item.testcase_type.name == imported_testrun['testcase'].get('testcase_type')
    # teststep sequence
    assert len(item.teststep_sequences) == 1
    item = item.teststep_sequences[0]
    assert item.name == imported_testrun['teststep_sequence'].get('name')
    assert item.description == imported_testrun['teststep_sequence'].get('description_import')
    assert item.classname.name == imported_testrun['teststep_sequence'].get('classname')
    # teststeps
    assert len(item.teststeps) == 5
    for ts_index, teststep, teststep_ref in zip(range(1,100), item.teststeps, imported_testrun.get('teststeps')):
        assert teststep.name == teststep_ref.get('name'), f'TestStep {ts_index}: name does not match'
        assert teststep.description == teststep_ref.get('description_import'), f'TestStep {ts_index}: description does not match'
        assert (teststep.locator_type.name if teststep.locator_type else teststep.locator_type) == teststep_ref.get('locator_type'), \
            f'TestStep {ts_index}: locator type does not match'
        assert teststep.locator == teststep_ref.get('locator'), f'TestStep {ts_index}: locator does not match'
        assert teststep.value == teststep_ref.get('value'), f'TestStep {ts_index}: value does not match'
        assert teststep.value2 == teststep_ref.get('value2'), f'TestStep {ts_index}: value2 does not match'
        assert teststep.comparison == teststep_ref.get('comparison'), f'TestStep {ts_index}: comparison does not match'
        assert teststep.timeout == teststep_ref.get('timeout'), f'TestStep {ts_index}: timeout does not match'
        assert teststep.optional == teststep_ref.get('optional'), f'TestStep {ts_index}: optional does not match'
        assert teststep.release == teststep_ref.get('release'), f'TestStep {ts_index}: release does not match'


def test_update_lazy_simple(client, test_user, default_supports, db_supports):
    # url
    url = '/testrun/import'
    db_supports(5)

    # import testrun
    login(client, test_user)
    with open(os.path.join(path_to_data, 'testrun_simple.xlsx'), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data)

    # lazy update
    with open(os.path.join(path_to_data, 'testrun_simple.xlsx'), 'rb') as f:
        file_data['file'] = f
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import Testrun. Status code: {r.status_code}'
    # assert success message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-success")]')), 'Success message is absent for CORRECT update'

    # assert updated testrun
    with open(os.path.join(path_to_data, 'testrun_simple.json'), 'r') as f:
        imported_testrun = json.load(f)
    # testrun
    item = models.Testrun.query.filter_by(name=imported_testrun['testrun'].get('name')).first()
    assert bool(item), 'Imported Testrun NOT saved to DB'
    assert item.description == imported_testrun['testrun'].get('description_update')
    # testcase sequence
    assert len(item.testcase_sequences) == 1
    item = item.testcase_sequences[0]
    assert item.name == imported_testrun['testcase_sequence'].get('name')
    assert item.description == imported_testrun['testcase_sequence'].get('description_update')
    assert item.classname.name == imported_testrun['testcase_sequence'].get('classname')
    # testcase
    assert len(item.testcases) == 1
    item = item.testcases[0]
    assert item.name == imported_testrun['testcase'].get('name')
    assert item.description == imported_testrun['testcase'].get('description_update')
    assert item.classname.name == imported_testrun['testcase'].get('classname')
    assert item.browser_type.name == imported_testrun['testcase'].get('browser_type')
    assert item.testcase_type.name == imported_testrun['testcase'].get('testcase_type')
    # teststep sequence
    assert len(item.teststep_sequences) == 1
    item = item.teststep_sequences[0]
    assert item.name == imported_testrun['teststep_sequence'].get('name')
    assert item.description == imported_testrun['teststep_sequence'].get('description_update')
    assert item.classname.name == imported_testrun['teststep_sequence'].get('classname')
    # teststeps
    assert len(item.teststeps) == 5
    for ts_index, teststep, teststep_ref in zip(range(1,100), item.teststeps, imported_testrun.get('teststeps')):
        assert teststep.name == teststep_ref.get('name'), f'TestStep {ts_index}: name does not match'
        assert teststep.description == teststep_ref.get('description_update'), \
            f'TestStep {ts_index}: description does not match'
        assert (teststep.locator_type.name if teststep.locator_type else teststep.locator_type) == teststep_ref.get('locator_type'), \
            f'TestStep {ts_index}: locator type does not match' 
        assert teststep.locator == teststep_ref.get('locator'), f'TestStep {ts_index}: locator does not match'
        assert teststep.value == teststep_ref.get('value'), f'TestStep {ts_index}: value does not match'
        assert teststep.value2 == teststep_ref.get('value2'), f'TestStep {ts_index}: value2 does not match'
        assert teststep.comparison == teststep_ref.get('comparison'), f'TestStep {ts_index}: comparison does not match'
        assert teststep.timeout == teststep_ref.get('timeout'), f'TestStep {ts_index}: timeout does not match'
        assert teststep.optional == teststep_ref.get('optional'), f'TestStep {ts_index}: optional does not match'
        assert teststep.release == teststep_ref.get('release'), f'TestStep {ts_index}: release does not match'


def test_import_complex(client, test_user, default_supports, db_supports):
    # set-up 
    url='/testrun/import'
    db_supports(10)
    login(client, test_user)

    # import testrun
    with open(os.path.join(path_to_data, 'testrun_complex.xlsx'), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import Testrun. Status code: {r.status_code}'
    # assert success message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-success")]')), 'Success message is absent for CORRECT import'

    # assert imported testrun
    with open(os.path.join(path_to_data, 'testrun_complex.json'), 'r') as f:
        reference_data = json.load(f)
    tr_ref = reference_data.get('testrun')
    descriptions = reference_data.get('descriptions')
    # testrun
    tr = models.Testrun.query.filter_by(name=tr_ref.get('name')).first()
    assert bool(tr), 'Imported Testrun NOT saved to DB'
    assert tr.description == descriptions.get('import')
    # testcase sequence
    assert len(tr.testcase_sequences) == len(tr_ref.get('testcase_sequences'))
    for tcs_index, tcs, tcs_ref in zip(range(1,100), tr.testcase_sequences, tr_ref.get('testcase_sequences')):
        assert tcs.name == tcs_ref.get('name'), f'TestCaseSequence {tcs_index}: name does not match'
        assert tcs.description == descriptions.get('import'), f'TestCaseSequence {tcs_index}: description does not match'
        assert tcs.classname.name == tcs_ref.get('classname'), f'TestCaseSequence {tcs_index}: classname does not match'
        # testcase
        assert len(tcs.testcases) == len(tcs_ref.get('testcases')), f'TestCaseSequence {tcs_index}: namber of children does not match'
        for tc_index, tc, tc_ref in zip(range(1,100), tcs.testcases, tcs_ref.get('testcases')):
            assert tc.name == tc_ref.get('name'), f'TestCase {tc_index}: name does not match'
            assert tc.description == descriptions.get('import'), f'TestCase {tc_index}: description does not match'
            assert tc.classname.name == tc_ref.get('classname'), f'TestCase {tc_index}: classname does not match'
            assert tc.browser_type.name == tc_ref.get('browser_type'), f'TestCase {tc_index}: browser type does not match'
            assert tc.testcase_type.name == tc_ref.get('testcase_type'), f'TestCase {tc_index}: testcase type does not match'
            # teststep sequence
            assert len(tc.teststep_sequences) == len(tc_ref.get('teststep_sequences')), f'TestCase {tc_index}: number of children does not match'
            for tss_index, tss, tss_ref in zip(range(1,100), tc.teststep_sequences, tc_ref.get('teststep_sequences')):
                assert tss.name == tss_ref.get('name'), f'TestStepSequence {tss_index}: name does not match'
                assert tss.description == descriptions.get('import'), f'TestStepSequence {tss_index}: description does not match'
                assert tss.classname.name == tss_ref.get('classname'), f'TestStepSequence {tss_index}: classname does not match'
                # teststeps
                assert len(tss.teststeps) == len(tss_ref.get('teststeps')), f'TestStepSequence {tss_index}: number of children does not match'
                for ts_index, ts, ts_ref in zip(range(1,100), tss.teststeps, tss_ref.get('teststeps')):
                    assert ts.name == ts_ref.get('name'), f'TestStep {ts_index}: name does not match'
                    assert ts.description == descriptions.get('import'), f'TestStep {ts_index}: description does not match'
                    assert (ts.locator_type.name if ts.locator_type else ts.locator_type) == ts_ref.get('locator_type'), \
                        f'TestStep {ts_index}: locator type does not match'
                    assert ts.locator == ts_ref.get('locator'), f'TestStep {ts_index}: locator does not match'
                    assert ts.value == ts_ref.get('value'), f'TestStep {ts_index}: value does not match'
                    assert ts.value2 == ts_ref.get('value2'), f'TestStep {ts_index}: value2 does not match'
                    assert ts.comparison == ts_ref.get('comparison'), f'TestStep {ts_index}: comparison does not match'
                    assert ts.timeout == ts_ref.get('timeout'), f'TestStep {ts_index}: timeout does not match'
                    assert ts.optional == ts_ref.get('optional'), f'TestStep {ts_index}: optional does not match'
                    assert ts.release == ts_ref.get('release'), f'TestStep {ts_index}: release does not match'


def test_update_complex_lazy(client, test_user, default_supports, db_supports):
    # set-up 
    url='/testrun/import'
    db_supports(10)
    login(client, test_user)

    # import testrun
    with open(os.path.join(path_to_data, 'testrun_complex.xlsx'), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data, follow_redirects=True)

    # lazy update
    with open(os.path.join(path_to_data, 'testrun_complex.xlsx'), 'rb') as f:
        file_data['file'] = f
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import Testrun. Status code: {r.status_code}'
    # assert success message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-success")]')), 'Success message is absent for CORRECT import'

    # assert imported testrun
    with open(os.path.join(path_to_data, 'testrun_complex.json'), 'r') as f:
        reference_data = json.load(f)
    tr_ref = reference_data.get('testrun')
    descriptions = reference_data.get('descriptions')
    # testrun
    tr = models.Testrun.query.filter_by(name=tr_ref.get('name')).first()
    assert bool(tr), 'Imported Testrun NOT saved to DB'
    assert tr.description == descriptions.get('update')
    # testcase sequence
    assert len(tr.testcase_sequences) == len(tr_ref.get('testcase_sequences'))
    for tcs_index, tcs, tcs_ref in zip(range(1,100), tr.testcase_sequences, tr_ref.get('testcase_sequences')):
        assert tcs.name == tcs_ref.get('name'), f'TestCaseSequence {tcs_index}: name does not match'
        assert tcs.description == descriptions.get('update'), f'TestCaseSequence {tcs_index}: description does not match'
        assert tcs.classname.name == tcs_ref.get('classname'), f'TestCaseSequence {tcs_index}: classname does not match'
        # testcase
        assert len(tcs.testcases) == len(tcs_ref.get('testcases')), f'TestCaseSequence {tcs_index}: namber of children does not match'
        for tc_index, tc, tc_ref in zip(range(1,100), tcs.testcases, tcs_ref.get('testcases')):
            assert tc.name == tc_ref.get('name'), f'TestCase {tc_index}: name does not match'
            assert tc.description == descriptions.get('update'), f'TestCase {tc_index}: description does not match'
            assert tc.classname.name == tc_ref.get('classname'), f'TestCase {tc_index}: classname does not match'
            assert tc.browser_type.name == tc_ref.get('browser_type'), f'TestCase {tc_index}: browser type does not match'
            assert tc.testcase_type.name == tc_ref.get('testcase_type'), f'TestCase {tc_index}: testcase type does not match'
            # teststep sequence
            assert len(tc.teststep_sequences) == len(tc_ref.get('teststep_sequences')), f'TestCase {tc_index}: number of children does not match'
            for tss_index, tss, tss_ref in zip(range(1,100), tc.teststep_sequences, tc_ref.get('teststep_sequences')):
                assert tss.name == tss_ref.get('name'), f'TestStepSequence {tss_index}: name does not match'
                assert tss.description == descriptions.get('update'), f'TestStepSequence {tss_index}: description does not match'
                assert tss.classname.name == tss_ref.get('classname'), f'TestStepSequence {tss_index}: classname does not match'
                # teststeps
                assert len(tss.teststeps) == len(tss_ref.get('teststeps')), f'TestStepSequence {tss_index}: number of children does not match'
                for ts_index, ts, ts_ref in zip(range(1,100), tss.teststeps, tss_ref.get('teststeps')):
                    assert ts.name == ts_ref.get('name'), f'TestStep {ts_index}: name does not match'
                    assert ts.description == descriptions.get('update'), f'TestStep {ts_index}: description does not match'
                    assert (ts.locator_type.name if ts.locator_type else ts.locator_type) == ts_ref.get('locator_type'), \
                        f'TestStep {ts_index}: locator type does not match'
                    assert ts.locator == ts_ref.get('locator'), f'TestStep {ts_index}: locator does not match'
                    assert ts.value == ts_ref.get('value'), f'TestStep {ts_index}: value does not match'
                    assert ts.value2 == ts_ref.get('value2'), f'TestStep {ts_index}: value2 does not match'
                    assert ts.comparison == ts_ref.get('comparison'), f'TestStep {ts_index}: comparison does not match'
                    assert ts.timeout == ts_ref.get('timeout'), f'TestStep {ts_index}: timeout does not match'
                    assert ts.optional == ts_ref.get('optional'), f'TestStep {ts_index}: optional does not match'
                    assert ts.release == ts_ref.get('release'), f'TestStep {ts_index}: release does not match'


def test_import_json(client, test_user, default_supports, db_supports):
    # set-up 
    url='/testrun/import'
    db_supports(10)
    login(client, test_user)

    # import testrun
    with open(os.path.join(path_to_data, 'testrun_definition.json'), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import Testrun. Status code: {r.status_code}'
    # assert success message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-success")]')), 'Success message is absent for CORRECT import'

    # assert imported testrun
    with open(os.path.join(path_to_data, 'testrun_complex.json'), 'r') as f:
        reference_data = json.load(f)
    tr_ref = reference_data.get('testrun')
    descriptions = reference_data.get('descriptions')
    # testrun
    tr = models.Testrun.query.filter_by(name=tr_ref.get('name_json')).first()
    assert bool(tr), 'Imported Testrun NOT saved to DB'
    assert tr.description == descriptions.get('json')
    # testcase sequence
    assert len(tr.testcase_sequences) == len(tr_ref.get('testcase_sequences'))
    for tcs_index, tcs, tcs_ref in zip(range(1,100), tr.testcase_sequences, tr_ref.get('testcase_sequences')):
        assert tcs.name == tcs_ref.get('name_json'), f'TestCaseSequence {tcs_index}: name does not match'
        assert tcs.description == descriptions.get('json'), f'TestCaseSequence {tcs_index}: description does not match'
        assert tcs.classname.name == tcs_ref.get('classname'), f'TestCaseSequence {tcs_index}: classname does not match'
        # testcase
        assert len(tcs.testcases) == len(tcs_ref.get('testcases')), f'TestCaseSequence {tcs_index}: namber of children does not match'
        for tc_index, tc, tc_ref in zip(range(1,100), tcs.testcases, tcs_ref.get('testcases')):
            assert tc.name == tc_ref.get('name_json'), f'TestCase {tc_index}: name does not match'
            assert tc.description == descriptions.get('json'), f'TestCase {tc_index}: description does not match'
            assert tc.classname.name == tc_ref.get('classname'), f'TestCase {tc_index}: classname does not match'
            assert tc.browser_type.name == tc_ref.get('browser_type'), f'TestCase {tc_index}: browser type does not match'
            assert tc.testcase_type.name == tc_ref.get('testcase_type'), f'TestCase {tc_index}: testcase type does not match'
            # teststep sequence
            assert len(tc.teststep_sequences) == len(tc_ref.get('teststep_sequences')), f'TestCase {tc_index}: number of children does not match'
            for tss_index, tss, tss_ref in zip(range(1,100), tc.teststep_sequences, tc_ref.get('teststep_sequences')):
                assert tss.name == tss_ref.get('name_json'), f'TestStepSequence {tss_index}: name does not match'
                assert tss.description == descriptions.get('json'), f'TestStepSequence {tss_index}: description does not match'
                assert tss.classname.name == tss_ref.get('classname'), f'TestStepSequence {tss_index}: classname does not match'
                # teststeps
                assert len(tss.teststeps) == len(tss_ref.get('teststeps')), f'TestStepSequence {tss_index}: number of children does not match'
                for ts_index, ts, ts_ref in zip(range(1,100), tss.teststeps, tss_ref.get('teststeps')):
                    assert ts.name == ts_ref.get('name_json'), f'TestStep {ts_index}: name does not match'
                    assert ts.description == descriptions.get('json'), f'TestStep {ts_index}: description does not match'
                    assert (ts.locator_type.name if ts.locator_type else ts.locator_type) == ts_ref.get('locator_type'), \
                        f'TestStep {ts_index}: locator type does not match'
                    assert ts.locator == ts_ref.get('locator'), f'TestStep {ts_index}: locator does not match'
                    assert ts.value == ts_ref.get('value'), f'TestStep {ts_index}: value does not match'
                    assert ts.value2 == ts_ref.get('value2'), f'TestStep {ts_index}: value2 does not match'
                    assert ts.comparison == ts_ref.get('comparison'), f'TestStep {ts_index}: comparison does not match'
                    assert ts.timeout == ts_ref.get('timeout'), f'TestStep {ts_index}: timeout does not match'
                    assert ts.optional == ts_ref.get('optional'), f'TestStep {ts_index}: optional does not match'
                    assert ts.release == ts_ref.get('release'), f'TestStep {ts_index}: release does not match'

#
# Update Testrun
#

def test_update_form(client, test_user):
    # login
    url = f'/testrun'
    login(client, test_user)
    r = client.get(url)

    # assert page elements
    page = html.fromstring(r.data)  
    form = page.xpath('//form[@id="updateForm"]')
    assert bool(form), f'Update Form is absent'
    form = form[0]

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field is absent'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # file field
    field = form.xpath('.//input[@id="file"]')
    assert bool(field), 'File field is absent'
    value = bool(field[0].attrib.get('value'))
    assert not value, f'Name field is NOT empty'
    
    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))


@pytest.mark.parametrize('count', [(1,1), (3,5)])
def test_update_simple(client, test_user, default_supports, db_supports, populate):
    # set-up
    db_supports(5)    
    login(client, test_user)
    
    # get testrun from db
    item = choice(models.Testrun.query.all())

    # update
    url = f'/testrun/{item.uuid}/import'
    testrun_src = 'testrun_simple.xlsx'
    with open(os.path.join(path_to_data, testrun_src), 'rb') as f:
        file_data = { 
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import Testrun. Status code: {r.status_code}'
    # assert success message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-success")]')), 'Success message is absent for CORRECT import'

    # assert updated testrun
    with open(os.path.join(path_to_data, 'testrun_simple.json'), 'r') as f:
        imported_testrun = json.load(f)
    # testrun
    item = models.Testrun.query.get(item.id)
    assert bool(item), 'Imported Testrun NOT saved to DB'
    assert (item.description == imported_testrun['testrun'].get('description_update')) or \
        (item.description == imported_testrun['testrun'].get('description_import'))
    # testcase sequence
    assert len(item.testcase_sequences) == 1
    item = item.testcase_sequences[0]
    assert (item.description == imported_testrun['testcase_sequence'].get('description_update')) or \
        (item.description == imported_testrun['testcase_sequence'].get('description_import')) 
    assert item.classname.name == imported_testrun['testcase_sequence'].get('classname')
    # testcase
    assert len(item.testcases) == 1
    item = item.testcases[0]
    assert (item.description == imported_testrun['testcase'].get('description_update')) or \
        (item.description == imported_testrun['testcase'].get('description_import'))
    assert item.classname.name == imported_testrun['testcase'].get('classname')
    assert item.browser_type.name == imported_testrun['testcase'].get('browser_type')
    assert item.testcase_type.name == imported_testrun['testcase'].get('testcase_type')
    # teststep sequence
    assert len(item.teststep_sequences) == 1
    item = item.teststep_sequences[0]
    assert (item.description == imported_testrun['teststep_sequence'].get('description_update')) or \
        (item.description == imported_testrun['teststep_sequence'].get('description_import'))
    assert item.classname.name == imported_testrun['teststep_sequence'].get('classname')
    # teststeps
    assert len(item.teststeps) == 5
    for teststep, teststep_ref in zip(item.teststeps, imported_testrun.get('teststeps')):
        assert (teststep.description == teststep_ref.get('description_update')) or \
             (teststep.description == teststep_ref.get('description_import'))
        assert (teststep.locator_type.name if teststep.locator_type else teststep.locator_type) == teststep_ref.get('locator_type') 
        assert teststep.locator == teststep_ref.get('locator')
        assert teststep.value == teststep_ref.get('value')
        assert teststep.value2 == teststep_ref.get('value2')
        assert teststep.comparison == teststep_ref.get('comparison')
        assert teststep.timeout == teststep_ref.get('timeout')
        assert teststep.optional == teststep_ref.get('optional')
        assert teststep.release == teststep_ref.get('release')


@pytest.mark.parametrize('count', [(1,1), (3,5)])   
def test_update_complex(client, test_user, default_supports, db_supports, populate):
    # set-up 
    db_supports(10)
    login(client, test_user)

    # get random testrun from db
    tr = choice(models.Testrun.query.all())

    # update testrun
    url = f'/testrun/{tr.uuid}/import'
    with open(os.path.join(path_to_data, 'testrun_complex.xlsx'), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import Testrun. Status code: {r.status_code}'
    # assert success message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-success")]')), 'Success message is absent for CORRECT import'

    # assert imported testrun
    with open(os.path.join(path_to_data, 'testrun_complex.json'), 'r') as f:
        reference_data = json.load(f)
    tr_ref = reference_data.get('testrun')
    descriptions = reference_data.get('descriptions')

    # update testrun
    tr = models.Testrun.query.get(tr.id)
    assert bool(tr), 'Imported Testrun NOT saved to DB'
    assert tr.description == descriptions.get('update')
    # testcase sequence
    assert len(tr.testcase_sequences) == len(tr_ref.get('testcase_sequences'))
    for tcs_index, tcs, tcs_ref in zip(range(1,100), tr.testcase_sequences, tr_ref.get('testcase_sequences')):
        assert tcs.description == descriptions.get('update') or tcs.description == descriptions.get('import'), \
            f'TestCaseSequence {tcs_index}: description does not match'
        assert tcs.classname.name == tcs_ref.get('classname'), f'TestCaseSequence {tcs_index}: classname does not match'
        # testcase
        assert len(tcs.testcases) == len(tcs_ref.get('testcases')), f'TestCaseSequence {tcs_index}: namber of children does not match'
        for tc_index, tc, tc_ref in zip(range(1,100), tcs.testcases, tcs_ref.get('testcases')):
            assert tc.description == descriptions.get('update') or tc.description == descriptions.get('import'), \
                f'TestCase {tc_index}: description does not match'
            assert tc.classname.name == tc_ref.get('classname'), f'TestCase {tc_index}: classname does not match'
            assert tc.browser_type.name == tc_ref.get('browser_type'), f'TestCase {tc_index}: browser type does not match'
            assert tc.testcase_type.name == tc_ref.get('testcase_type'), f'TestCase {tc_index}: testcase type does not match'
            # teststep sequence
            assert len(tc.teststep_sequences) == len(tc_ref.get('teststep_sequences')), f'TestCase {tc_index}: number of children does not match'
            for tss_index, tss, tss_ref in zip(range(1,100), tc.teststep_sequences, tc_ref.get('teststep_sequences')):
                assert tss.description == descriptions.get('update') or tss.description == descriptions.get('import'), \
                    f'TestStepSequence {tss_index}: description does not match'
                assert tss.classname.name == tss_ref.get('classname'), f'TestStepSequence {tss_index}: classname does not match'
                # teststeps
                assert len(tss.teststeps) == len(tss_ref.get('teststeps')), f'TestStepSequence {tss_index}: number of children does not match'
                for ts_index, ts, ts_ref in zip(range(1,100), tss.teststeps, tss_ref.get('teststeps')):
                    assert ts.description == descriptions.get('update') or ts.description ==  descriptions.get('import'), \
                        f'TestStep {ts_index}: description does not match'
                    assert (ts.locator_type.name if ts.locator_type else ts.locator_type) == ts_ref.get('locator_type'), \
                        f'TestStep {ts_index}: locator type does not match'
                    assert ts.locator == ts_ref.get('locator'), f'TestStep {ts_index}: locator does not match'
                    assert ts.value == ts_ref.get('value'), f'TestStep {ts_index}: value does not match'
                    assert ts.value2 == ts_ref.get('value2'), f'TestStep {ts_index}: value2 does not match'
                    assert ts.comparison == ts_ref.get('comparison'), f'TestStep {ts_index}: comparison does not match'
                    assert ts.timeout == ts_ref.get('timeout'), f'TestStep {ts_index}: timeout does not match'
                    assert ts.optional == ts_ref.get('optional'), f'TestStep {ts_index}: optional does not match'
                    assert ts.release == ts_ref.get('release'), f'TestStep {ts_index}: release does not match'


#
# Import Fake Testrun
#
def test_import_fake(client, test_user):
    # url
    url = '/testrun/import'
    fake_testrun = 'testrun_fake.xlsx'

    # authorized GET access
    login(client, test_user)

    # import fake Testrun
    with open(os.path.join(path_to_data, fake_testrun), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import fake Testrun. Status code: {r.status_code}'
    # assert error message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-danger")]')
    assert bool(msg), 'Error message is absent'
    assert fake_testrun in msg[0].text, 'Error message does NOT contain exported file name'

@pytest.mark.parametrize('count', [(1,1)])
def test_update_fake(client, test_user, populate):
    # set-up
    item = models.Testrun.query.first()
    url = f'/testrun/{item.uuid}/import'
    login(client, test_user)
    
    # update with fake testrun
    fake_testrun = 'testrun_fake.xlsx'
    with open(os.path.join(path_to_data, fake_testrun), 'rb') as f:
        file_data = { 
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import fake Testrun. Status code: {r.status_code}'
    # assert error message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-danger")]')
    assert bool(msg), 'Error message is absent'
    assert fake_testrun in msg[0].text, 'Error message does NOT contain exported file name'

